
/*******************************************************************************
*
* File time1.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of Zhat
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "sw_term.h"
#include "dfl.h"
#include "global.h"


static void random_basis(int Ns)
{
   int i;
   spinor **ws;

   ws=reserve_ws(Ns);

   for (i=0;i<Ns;i++)
      random_s(VOLUME,ws[i],1.0f);
   
   dfl_subspace(ws);

   release_ws();
}


int main(int argc,char *argv[])
{
   int my_rank,nflds;
   int bs[4],Ns,nb,isw,nbb,nv;
   int n,i,count;
   double wt1,wt2,wdt;
   complex **wv;
   zop_t *zop;
   FILE *fin=NULL,*flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("time1.log","w",stdout);
      fin=freopen("time1.in","r",stdin);
      
      printf("\n");
      printf("Timing of Zhat\n");
      printf("--------------\n\n");   
      
      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("Ns","%d",&Ns);
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("Ns = %d\n\n",Ns);
      fflush(flog);      
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   

   start_ranlux(0,12345);
   geometry();

   alloc_ud();
   alloc_u();
   alloc_swd();
   alloc_sw();
   alloc_s(Ns);

   set_lat_parms(0.0,0.12,1.0);
   set_dfl_parms(bs,Ns,4,1,1,1,1,1,1,0.12,1.0e-3,1.0e-6,1.0e-12,0,0);
   
   random_ud();
   assign_ud2u();
   sw_term();
   assign_swd2sw();

   blk_list(DFL_BLOCKS,&nb,&isw);
   nv=nb*Ns;

   if (my_rank==0)
   {
      printf("Number of points = %d\n",VOLUME);
      printf("Number of blocks = %d\n",nb);
      printf("Number of points/block = %d\n",bs[0]*bs[1]*bs[2]*bs[3]);
      printf("Vector field size = %.2f KB\n",
             (double)(sizeof(complex)*nb*Ns)*1.0e-3);
      printf("Zhat array size = %.2f MB\n\n",
             (double)(sizeof(complex)*8*nb*Ns*Ns)*1.0e-6);
   }

   nflds=(int)(4.0e6/(double)(sizeof(complex)*nb*Ns));
   if ((nflds%2)!=0)
      nflds+=1;
   if (nflds==0)
      nflds=2;

   set_dfl_parms(bs,Ns,4,1,1,1,1,1,1,0.12,1.0e-3,1.0e-6,1.0e-12,nflds,0);   
   
   random_basis(Ns);
   wv=reserve_wv(nflds);

   for (i=0;i<nflds;i++)
      random_v(nv,wv[i],1.0f);

   zop=zophat();
   nbb=(*zop).nbb;

   n=(int)(1.0e3/(double)(8*Ns));
   if (n<1)
      n=1;
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime(); 
   for (count=0;count<n;count++)
   {
      for (i=0;i<nflds;i+=2)
         Zhat(wv[i],wv[i+1]);
   }
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   
   wdt=2.0e6*(wt2-wt1)/(double)(n*nflds);

   if (my_rank==0)
   {   
      printf("Time per application of Zhat, including communications:\n");
      printf("Time per block: %4.3f micro sec ",wdt/(double)(nb));
      printf("(%d Mflops [%d bit arithmetic])\n",
             (int)(64.0*(double)(nb*Ns*Ns)/wdt),(int)(4*sizeof(complex)));
      printf("Time per point: %4.3f micro sec\n\n",wdt/(double)(VOLUME));
   }

   if (NPROC>1)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime(); 
      for (count=0;count<n;count++)
      {
         for (i=0;i<nflds;i+=2)
         {
            set_v2zero((nbb/2)*Ns,wv[i+1]+nb*Ns);         
            cpv_int_bnd(wv[i]);
            cpv_ext_bnd(wv[i+1]);
         }
      }
      MPI_Barrier(MPI_COMM_WORLD);
      wt2=MPI_Wtime();
      wdt=2.0e6*(wt2-wt1)/(double)(n*nflds);      

      if (my_rank==0)
      {
         printf("There are %d boundary blocks\n",nbb);
         printf("Time per application of Zhat for the communications:\n");
         printf("Time per block: %4.3f micro sec\n",wdt/(double)(nb));
         printf("Time per point: %4.3f micro sec\n\n",wdt/(double)(VOLUME));
      }
   }
   
   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();   
   exit(0);
}
