
/*******************************************************************************
*
* File read3b.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Reads and evaluates data from the *.dat files created by the program run3.
* The file to be read has to be specified on the command line
*
* This program prints the probability that |dH|>r, for r=1,2,10,100,1000,
* to stdout
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "start.h"
#include "misc.h"
#include "extras.h"

typedef struct
{
   int nt,ns,iac;
   int stat_rop,stat_frf,stat_acf;
   int stat_gcr_frf[3],stat_gcr_acf[3];
   int stat_dfl_mds,stat_dfl_upd[2];
   double dact,avpl;
} dat_t;

static int nms,nfirst,nlast,neff;
static dat_t *adat;


static int read_dat(int n,dat_t *ndat,FILE *fin)
{
   int i,iw,ic,ie;
   stdint_t istd[15];
   double dstd[2];

   ic=0;
   ie=endianness();
   
   for (i=0;i<n;i++)
   {
      iw =fread(istd,sizeof(stdint_t),15,fin);
      iw+=fread(dstd,sizeof(double),2,fin);

      if (iw==17)
         ic+=1;
      else
         break;

      if (ie==BIG_ENDIAN)
      {
         bswap_int(15,istd);
         bswap_double(2,dstd);
      }
      
      (*ndat).nt=(int)(istd[0]);
      (*ndat).ns=(int)(istd[1]);
      (*ndat).iac=(int)(istd[2]);

      (*ndat).stat_rop=(int)(istd[3]);
      (*ndat).stat_frf=(int)(istd[4]);
      (*ndat).stat_acf=(int)(istd[5]);       

      (*ndat).stat_gcr_frf[0]=(int)(istd[ 6]);
      (*ndat).stat_gcr_frf[1]=(int)(istd[ 7]);
      (*ndat).stat_gcr_frf[2]=(int)(istd[ 8]);      

      (*ndat).stat_gcr_acf[0]=(int)(istd[ 9]);
      (*ndat).stat_gcr_acf[1]=(int)(istd[10]);
      (*ndat).stat_gcr_acf[2]=(int)(istd[11]);       

      (*ndat).stat_dfl_mds=(int)(istd[12]);
      (*ndat).stat_dfl_upd[0]=(int)(istd[13]);
      (*ndat).stat_dfl_upd[1]=(int)(istd[14]);      
      
      (*ndat).dact=dstd[0];
      (*ndat).avpl=dstd[1];

      ndat+=1;
   }
   
   return ic;
} 


static void read_file(char *file)
{
   int n,ir;
   long ipos;
   dat_t ndat;
   FILE *fdat;

   fdat=fopen(file,"rb");
   error(fdat==NULL,1,"read_file [read3b.c]","Unable to open data file");

   printf("Read data from file %s\n\n",file);
   ipos=ftell(fdat);
   ir=0;
      
   for (n=0;feof(fdat)==0;n++)
      ir+=read_dat(1,&ndat,fdat);

   error((ir!=(n-1))||(ferror(fdat)!=0),1,"read_file [read3b.c]",
         "Incorrect read count or read error");   
   
   nms=n-1;
   adat=amalloc(nms*sizeof(*adat),3);
   error(adat==NULL,1,"read_file [read3b.c]",
         "Unable to allocate data array");

   fseek(fdat,ipos,SEEK_SET);
   ir=read_dat(nms,adat,fdat);

   error((ir!=nms)||(feof(fdat)!=0)||(ferror(fdat)!=0),1,"read_file [read3b.c]",
         "Error while reading data file");
   fclose(fdat);
}


static void select_range(void)
{
   int n,nf,nl,nt,np,dn,ie;
   
   printf("There are %d measurements (trajectories no %d - %d)\n",
          nms,adat[0].nt,adat[nms-1].nt);
   printf("Range [nfirst,nlast] of trajectories to analyse: ");
   scanf("%d %d",&nfirst,&nlast);

   nf=0;
   nl=0;
      
   for (n=0;n<nms;n++)
   {
      nt=adat[n].nt;
      
      if (nt<nfirst)
         nf+=1;

      if (nt<=nlast)
         nl+=1;
   }

   nfirst=nf;
   nlast=nl;
   neff=nlast-nfirst;
      
   printf("Keep %d measurements (trajectories no %d - %d)\n\n",
          neff,adat[nfirst].nt,adat[nlast-1].nt);

   if (neff>1)
   {
      np=adat[nfirst].nt;
      dn=adat[nfirst+1].nt-adat[nfirst].nt;

      if (dn<=0)
         ie=1;
      else
      {
         ie=0;
         
         for (n=(nfirst+1);n<nlast;n++)
         {
            nt=adat[n].nt;

            if ((nt-np)!=dn)
            {
               ie=2;
               break;
            }

            np=nt;
         }
      }

      error(ie!=0,1,"select_range [read3b.c]",
            "Irregular trajectory numbers between successive\n"
            "measurements in selected range");
   }
}


int main(int argc,char *argv[])
{
   int n,i;
   double r[5],prob[5],x,dx;
   char fmt[128];

   error(argc!=2,1,"main [read3a.c]","Syntax: read3a <filename>");

   printf("\n");
   printf("DD-HMC simulation of two-flavour O(a) improved QCD\n");
   printf("--------------------------------------------------\n\n");

   read_file(argv[1]);
   select_range();

   r[0]=1.0;
   r[1]=2.0;
   r[2]=10.0;
   r[3]=100.0;
   r[4]=1000.0;
   
   for (i=0;i<5;i++)
      prob[i]=0.0;
   
   for (n=nfirst;n<nlast;n++)
   {
      x=fabs(adat[n].dact);

      for (i=0;i<5;i++)
         if (x>r[i])
            prob[i]+=1.0;
   }

   printf("Energy deficits at the end of the trajectories:\n\n");
   
   for (i=0;i<5;i++)
   {
      x=prob[i]/(double)(neff);
      dx=sqrt((x-x*x)/(double)(neff));

      if (x>0.0)
      {
         n=digits(x,dx,"f");
         sprintf(fmt,"Prob(|dH|>%4d) =%%- .%df(%%2d)\n",(int)(r[i]),n);
         printf(fmt,x,(int)(exp(log(10.0)*(double)(n))*dx));
      }
      else
      {
         printf("Prob(|dH|>%4d) = 0.0\n",(int)(r[i]));
         break;
      }
   }

   printf("\n");
   exit(0);
}
