
/*******************************************************************************
*
* File time2.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the SU(3) x SU(3)-vector multiplication (single-precision programs)
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "su3.h"
#include "random.h"
#include "misc.h"
#include "start.h"

#define N 20000000

#if (defined SSE)

#include "sse.h"

static su3 u[4] __attribute__ ((aligned (16)));
static su3_vector s[8],r[8],t[8] __attribute__ ((aligned (16)));

#define _su3_fast_multiply(r1,r2,u,s1,s2) \
   _sse_pair_load(s1,s2); \
   _sse_su3_multiply(u); \
   _sse_pair_store_up(r1,r2)

#define _su3_fast_inverse_multiply(r1,r2,u,s1,s2) \
   _sse_pair_load(s1,s2); \
   _sse_su3_inverse_multiply(u); \
   _sse_pair_store_up(r1,r2)

#else

static su3 u[4];
static su3_vector s[8],r[8],t[8];

#endif


static void fast_multiply(void)
{
   _su3_fast_multiply((*(r  )),(*(r+1)),(*(u  )),(*(s  )),(*(s+1)));
   _su3_fast_multiply((*(r+2)),(*(r+3)),(*(u+1)),(*(s+2)),(*(s+3)));
   _su3_fast_multiply((*(r+4)),(*(r+5)),(*(u+2)),(*(s+4)),(*(s+5)));
   _su3_fast_multiply((*(r+6)),(*(r+7)),(*(u+3)),(*(s+6)),(*(s+7)));
}


static void fast_inverse_multiply(void)
{
   _su3_fast_inverse_multiply((*(t  )),(*(t+1)),(*(u  )),(*(r  )),(*(r+1)));
   _su3_fast_inverse_multiply((*(t+2)),(*(t+3)),(*(u+1)),(*(r+2)),(*(r+3)));
   _su3_fast_inverse_multiply((*(t+4)),(*(t+5)),(*(u+2)),(*(r+4)),(*(r+5)));
   _su3_fast_inverse_multiply((*(t+6)),(*(t+7)),(*(u+3)),(*(r+6)),(*(r+7)));
}


static void slow_multiply(void)
{
   _su3_multiply((*(t  )),(*(u  )),(*(s  )));
   _su3_multiply((*(t+1)),(*(u  )),(*(s+1)));
   _su3_multiply((*(t+2)),(*(u+1)),(*(s+2)));
   _su3_multiply((*(t+3)),(*(u+1)),(*(s+3)));
   _su3_multiply((*(t+4)),(*(u+2)),(*(s+4)));
   _su3_multiply((*(t+5)),(*(u+2)),(*(s+5)));
   _su3_multiply((*(t+6)),(*(u+3)),(*(s+6)));
   _su3_multiply((*(t+7)),(*(u+3)),(*(s+7)));
}


int main(void)
{
   int k,count;
   double t1,t2,dt;
   double delta,diff,norm;

   printf("\n");
   printf("Time per single-precision SU(3) x SU(3)-vector multiplication\n");
   printf("-------------------------------------------------------------\n\n");

   rlxs_init(0,123456);

   for (k=0;k<4;k++)
      random_su3(u+k);

   for (k=0;k<8;k++)
      gauss((float*)(s+k),6);

#if (defined SSE)

   t1=(double)clock();
   for (count=0;count<N;count++)
      fast_multiply();
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(8*N);

#if (defined SSE3)
      printf("Using SSE3 instructions:\n");
#elif (defined SSE2)
      printf("Using SSE2 instructions:\n");
#else
      printf("Using SSE instructions:\n");
#endif
   printf("The time per v=U*w is     %4.3f micro sec",dt);
   printf(" [%d Mflops]\n",(int)(66.0/dt));

   t1=(double)clock();
   for (count=0;count<N;count++)
      fast_inverse_multiply();
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(8*N);
   printf("The time per v=U^dag*w is %4.3f micro sec",dt);
   printf(" [%d Mflops]\n",(int)(66.0/dt));

   delta=0.0;

   for (k=0;k<8;k++)
   {
      _vector_sub_assign(t[k],s[k]);
      diff=(double)(_vector_prod_re(t[k],t[k]));
      norm=(double)(_vector_prod_re(s[k],s[k]));
      diff=sqrt(diff/norm);
      if (diff>delta)
         delta=diff;
   }

   printf("||w-U^dag*U*w||<= %.1e*||w||\n\n",delta);

#endif

   t1=(double)clock();
   for (count=0;count<(N/2);count++)
      slow_multiply();
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(4*N);

   printf("Using x86 FPU instructions:\n");
   printf("The time per v=U*w is     %4.3f micro sec",dt);
   printf(" [%d Mflops]\n",(int)(66.0/dt));

#if (defined SSE)

   delta=0.0;

   for (k=0;k<8;k++)
   {
      _vector_sub_assign(r[k],t[k]);
      diff=(double)(_vector_prod_re(r[k],r[k]));
      norm=(double)(_vector_prod_re(s[k],s[k]));
      diff=sqrt(diff/norm);
      if (diff>delta)
         delta=diff;
   }

   printf("||v_SSE-v_FPU||<= %.1e*||w||\n",delta);

#endif

   _su3_inverse_multiply((*(r  )),(*(u  )),(*(t  )));
   _su3_inverse_multiply((*(r+1)),(*(u  )),(*(t+1)));
   _su3_inverse_multiply((*(r+2)),(*(u+1)),(*(t+2)));
   _su3_inverse_multiply((*(r+3)),(*(u+1)),(*(t+3)));
   _su3_inverse_multiply((*(r+4)),(*(u+2)),(*(t+4)));
   _su3_inverse_multiply((*(r+5)),(*(u+2)),(*(t+5)));
   _su3_inverse_multiply((*(r+6)),(*(u+3)),(*(t+6)));
   _su3_inverse_multiply((*(r+7)),(*(u+3)),(*(t+7)));

   delta=0.0;

   for (k=0;k<8;k++)
   {
      _vector_sub_assign(r[k],s[k]);
      diff=(double)(_vector_prod_re(r[k],r[k]));
      norm=(double)(_vector_prod_re(s[k],s[k]));
      diff=sqrt(diff/norm);
      if (diff>delta)
         delta=diff;
   }

   printf("||w-U^dag*U*w||<= %.1e*||w||\n\n",delta);

   exit(0);
}

