
/*******************************************************************************
*
* File time3.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the SU(3) x SU(3)-vector multiplications (double-precision
* programs)
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "su3.h"
#include "random.h"
#include "misc.h"
#include "start.h"

#define N 30000000

#if (defined SSE2)

#include "sse2.h"

static su3_dble u[4] __attribute__ ((aligned (16)));
static su3_vector_dble s[4],r[4],t[4] __attribute__ ((aligned (16)));

#define _su3_fast_multiply(r,u,s) \
   _sse_load_dble(s); \
   _sse_su3_multiply_dble(u); \
   _sse_store_up_dble(r)

#define _su3_fast_inverse_multiply(r,u,s) \
   _sse_load_dble(s); \
   _sse_su3_inverse_multiply_dble(u); \
   _sse_store_up_dble(r)

#else

static su3_dble u[4];
static su3_vector_dble s[4],r[4],t[4];

#endif


static void fast_multiply(void)
{
   _su3_fast_multiply((*(r  )),(*(u  )),(*(s  )));
   _su3_fast_multiply((*(r+1)),(*(u+1)),(*(s+1)));
   _su3_fast_multiply((*(r+2)),(*(u+2)),(*(s+2)));
   _su3_fast_multiply((*(r+3)),(*(u+3)),(*(s+3)));
}


static void fast_inverse_multiply(void)
{
   _su3_fast_inverse_multiply((*(t  )),(*(u  )),(*(r  )));
   _su3_fast_inverse_multiply((*(t+1)),(*(u+1)),(*(r+1)));
   _su3_fast_inverse_multiply((*(t+2)),(*(u+2)),(*(r+2)));
   _su3_fast_inverse_multiply((*(t+3)),(*(u+3)),(*(r+3)));
}


static void slow_multiply(void)
{
   _su3_multiply((*(t  )),(*(u  )),(*(s  )));
   _su3_multiply((*(t+1)),(*(u+1)),(*(s+1)));
   _su3_multiply((*(t+2)),(*(u+2)),(*(s+2)));
   _su3_multiply((*(t+3)),(*(u+3)),(*(s+3)));
}


int main(void)
{
   int k,count;
   double t1,t2,dt;
   double delta,diff,norm;

   printf("\n");
   printf("Time per double-precision SU(3) x SU(3)-vector multiplication\n");
   printf("-------------------------------------------------------------\n\n");

   rlxd_init(1,123456);

   for (k=0;k<4;k++)
   {
      random_su3_dble(u+k);
      gauss_dble((double*)(s+k),6);
   }

#if (defined SSE2)

   t1=(double)clock();
   for (count=0;count<N;count++)
      fast_multiply();
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(4*N);

#if (defined SSE3)
      printf("Using SSE3 instructions:\n");
#else
      printf("Using SSE2 instructions:\n");
#endif
   printf("The time per v=U*w is     %4.3f micro sec",dt);
   printf(" [%d Mflops]\n",(int)(66.0/dt));

   t1=(double)clock();
   for (count=0;count<N;count++)
      fast_inverse_multiply();
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(4*N);

   printf("The time per v=U^dag*w is %4.3f micro sec",dt);
   printf(" [%d Mflops]\n",(int)(66.0/dt));

   delta=0.0;

   for (k=0;k<4;k++)
   {
      _vector_sub_assign(t[k],s[k]);
      diff=_vector_prod_re(t[k],t[k]);
      norm=_vector_prod_re(s[k],s[k]);
      diff=sqrt(diff/norm);
      if (diff>delta)
         delta=diff;
   }

   printf("||w-U^dag*U*w||<= %.1e*||w||\n\n",delta);

#endif

   t1=(double)clock();
   for (count=0;count<N;count++)
      slow_multiply();
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(4*N);

   printf("Using x86 FPU instructions:\n");
   printf("The time per v=U*w is     %4.3f micro sec",dt);
   printf(" [%d Mflops]\n",(int)(66.0/dt));

#if (defined SSE2)

   delta=0.0;

   for (k=0;k<4;k++)
   {
      _vector_sub_assign(r[k],t[k]);
      diff=_vector_prod_re(r[k],r[k]);
      norm=_vector_prod_re(s[k],s[k]);
      diff=sqrt(diff/norm);
      if (diff>delta)
         delta=diff;
   }

   printf("||v_SSE-v_FPU||<= %.1e*||w||\n",delta);

#endif

   _su3_inverse_multiply((*(r  )),(*(u  )),(*(t  )));
   _su3_inverse_multiply((*(r+1)),(*(u+1)),(*(t+1)));
   _su3_inverse_multiply((*(r+2)),(*(u+2)),(*(t+2)));
   _su3_inverse_multiply((*(r+3)),(*(u+3)),(*(t+3)));

   delta=0.0;

   for (k=0;k<4;k++)
   {
      _vector_sub_assign(r[k],s[k]);
      diff=_vector_prod_re(r[k],r[k]);
      norm=_vector_prod_re(s[k],s[k]);
      diff=sqrt(diff/norm);
      if (diff>delta)
         delta=diff;
   }

   printf("||w-U^dag*U*w||<= %.1e*||w||\n\n",delta);

   exit(0);
}

