
/*******************************************************************************
*
* File time2.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the Schwarz preconditioner
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "misc.h"
#include "start.h"
#include "update.h"
#include "linalg.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "sap_gcr.h"
#include "global.h"

static int my_rank,nmr,ncy;


static void Opz(spinor *pr,spinor *paz)
{
   spinor *rsm;

   rsm=pr+VOLUME;

   for (;pr<rsm;pr++)
   {
      _vector_sub((*paz).c1,(*pr).c1,(*paz).c1);
      _vector_sub((*paz).c2,(*pr).c2,(*paz).c2);
      _vector_sub((*paz).c3,(*pr).c3,(*paz).c3);
      _vector_sub((*paz).c4,(*pr).c4,(*paz).c4);
      paz+=1;
   }
}


static void Mj(int j,spinor *pr,spinor *pz,spinor *paz)
{
   int n,isz,isaz;

   isz=(pz-ps[0][0])/NSPIN;
   isaz=(paz-ps[0][0])/NSPIN;

   set_s2zero(VOLUME,pz);
   assign_s2s(VOLUME,pr,paz);

   for (n=0;n<ncy;n++)
      msap(nmr,isaz,isz);

   Opz(pr,paz);
}


int main(int argc,char *argv[])
{
   int n,count,bs[4];
   double wt1,wt2,wdt;
   spinor *pr,*pz,*paz;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("time2.log","w",stdout);
      fin=freopen("time2.in","r",stdin);

      printf("\n");
      printf("Timing of the Schwarz preconditioner\n");
      printf("------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      scanf("bs %d %d %d %d\n",&bs[0],&bs[1],&bs[2],&bs[3]);
      scanf("nmr %d ncy %d",&nmr,&ncy);
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("nmr = %d\n",nmr);
      printf("ncy = %d\n\n",ncy);
      fflush(flog);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy,1,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,12345);
   geometry();
#if (defined BGL)
   write_bglinfo(my_rank,"time2.bgl");
#endif
   alloc_ud();
   alloc_swd();

   alloc_s(3);
   alloc_bgr(GCR_BLOCKS,bs);

   set_lat_parms(0.0,0.1,0.2);
   random_ud();
   sw_term();
   assign_ud2ubgr(GCR_BLOCKS);
   n=assign_swd2swbgr(GCR_BLOCKS,ODD_PTS);
   error(n!=0,1,"main [time2.c]","Inversion of the SW term was not safe");

   pr=ps[0][0];
   pz=ps[1][0];
   paz=ps[2][0];

   n=2000000/(nmr*ncy*VOLUME);
   if (n<1)
      n=1;

   random_s(VOLUME,pr,1.0f);
   normalize(VOLUME,1,pr);
   Mj(0,pr,pz,paz);

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();
   for (count=0;count<n;count++)
      Mj(0,pr,pz,paz);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();

   if (my_rank==0)
   {
      wdt=1.0e6*(wt2-wt1)/(double)(n*VOLUME);
      printf("Time per lattice point: %.2f micro sec",wdt);
      printf(" (about %d Mflops [%d bit arithmetic])\n",
             (int)((double)(24+ncy*1944+ncy*nmr*2064)/wdt),
             (int)(sizeof(spinor))/3);
      printf("Time per lattice point and MR iteration: %.2f micro sec\n\n",
             wdt/(double)(ncy*nmr));
   }

   n*=nmr;
   nmr=0;

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();
   for (count=0;count<n;count++)
      Mj(0,pr,pz,paz);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();

   error_chk();

   if (my_rank==0)
   {
      wdt=wt2-wt1;
      printf("Without block solver (i.e. setting nmr=0):\n");
      wdt=1.0e6*(wt2-wt1)/(double)(n*VOLUME);
      printf("Time per lattice point: %.2f micro sec",wdt);
      printf(" (about %d Mflops [%d bit arithmetic])\n",
             (int)((double)(24+ncy*1944)/wdt),(int)(sizeof(spinor))/3);
      printf("Time per lattice point and Schwarz cycle: %.2f micro sec\n\n",
             wdt/(double)(ncy));
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
