
/*******************************************************************************
*
* File Zgen.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generic programs needed for the computation of the little Dirac operator
*
* The externally accessible functions are
*
*   void gather_sd(int vol,int *imb,spinor_dble *pk,spinor_dble *pl)
*     Assigns the spinor pk[imb[i]] to pl[i] (i=0,..,vol-1). The
*     output field must be different from the input spinor field 
*     (which is left unchanged)
*  
*   void apply_U2sd(int vol,int *imb,su3_dble *ud,spinor_dble *pk,
*                   spinor_dble *pl)
*     Multiplies the spinor pk[imb[i]] by the colour matrix ud[i] and
*     assigns the result to pl[i] (i=0,..,vol-1). The output field must
*     be different from the input spinor field (which is left unchanged)
*
*   void apply_Udag2sd(int vol,int *imb,su3_dble *ud,spinor_dble *pk,
*                      spinor_dble *pl)
*     Multiplies the spinor pk[imb[i]] by the adjoint of the colour
*     matrix ud[i] and assigns the result to pl[i] (i=0,..,vol-1). The
*     output field must be different from the input spinor field (which
*     is left unchanged)
*
* The following is an array of functions indexed by the direction mu=0,..,3
*
*   void (*spinor_prod_gamma[])(int vol,spinor_dble *pk,spinor_dble *pl,
*                               complex_dble *sp)
*      Computes the scalar products (pk,pl) and (pk,gamma_mu*pl) where
*      gamma_mu denotes the Dirac matrix with index mu and the spinor
*      fields are assumed to have vol elements. On exit the calculated
*      products are assigned to sp[0] and sp[1] respectively
*
* Notes:
*
* The representation of the Dirac matrices is the one specified in the notes
* "Implementation of the Dirac operator" which are included in the doc
* directory
*
* None of these programs involves communications. They are general purpose
* routines that know nothing about the underlying geometry. In particular,
* they can be called locally
*
* If SSE2 instructions are to be used, it is taken for granted  that the
* field arrays are aligned to 16 byte boundaries
*
*******************************************************************************/

#define ZGEN_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "misc.h"
#include "global.h"

#define MAX_LEVELS 8
#define BLK_LENGTH 16

static int cnt[MAX_LEVELS];

#if (defined SSE2)
#include "sse2.h"
static complex_dble sm0[MAX_LEVELS] __attribute__ ((aligned (16)));
static complex_dble sm1[MAX_LEVELS] __attribute__ ((aligned (16)));
static complex_dble z0,z1 __attribute__ ((aligned (16)));
static const sse_double sgn={1.0,-1.0};
#else
static int cnt[MAX_LEVELS];
static complex_dble sm0[MAX_LEVELS],sm1[MAX_LEVELS],z0,z1;
#endif


void gather_sd(int vol,int *imb,spinor_dble *pk,spinor_dble *pl)
{
   int *imm;

   imm=imb+vol;

   for (;imb<imm;imb++)
   {
      (*pl)=pk[*imb];
      pl+=1;
   }
}


#if (defined SSE2)

#define _zgen_su3_multiply(r,u,s) \
        _sse_load_dble(s); \
        _sse_su3_multiply_dble(u); \
        _sse_store_up_dble(r)

#define _zgen_su3_inverse_multiply(r,u,s) \
        _sse_load_dble(s); \
        _sse_su3_inverse_multiply_dble(u); \
        _sse_store_up_dble(r)


void apply_U2sd(int vol,int *imb,su3_dble *ud,spinor_dble *pk,
                spinor_dble *pl)
{
   int *imm;
   spinor_dble *s;

   imm=imb+vol;

   for (;imb<imm;)
   {
      s=pk+(*imb);
      imb+=1;

      if (imb<imm)
         _prefetch_spinor_dble(pk+(*imb));

      _zgen_su3_multiply((*pl).c1,(*ud),(*s).c1);
      _zgen_su3_multiply((*pl).c2,(*ud),(*s).c2);
      _zgen_su3_multiply((*pl).c3,(*ud),(*s).c3);
      _zgen_su3_multiply((*pl).c4,(*ud),(*s).c4);      

      ud+=1;
      pl+=1;
   }
}


void apply_Udag2sd(int vol,int *imb,su3_dble *ud,spinor_dble *pk,
                   spinor_dble *pl)
{
   int *imm;
   spinor_dble *s;

   imm=imb+vol;

   for (;imb<imm;)
   {
      s=pk+(*imb);
      imb+=1;

      if (imb<imm)
         _prefetch_spinor_dble(pk+(*imb));      

      _zgen_su3_inverse_multiply((*pl).c1,(*ud),(*s).c1);
      _zgen_su3_inverse_multiply((*pl).c2,(*ud),(*s).c2);
      _zgen_su3_inverse_multiply((*pl).c3,(*ud),(*s).c3);
      _zgen_su3_inverse_multiply((*pl).c4,(*ud),(*s).c4);      

      ud+=1;
      pl+=1;
   }
}  

#else

void apply_U2sd(int vol,int *imb,su3_dble *ud,spinor_dble *pk,
                spinor_dble *pl)
{
   int *imm;
   spinor_dble *s;

   imm=imb+vol;

   for (;imb<imm;imb++)
   {
      s=pk+(*imb);

      _su3_multiply((*pl).c1,(*ud),(*s).c1);
      _su3_multiply((*pl).c2,(*ud),(*s).c2);
      _su3_multiply((*pl).c3,(*ud),(*s).c3);
      _su3_multiply((*pl).c4,(*ud),(*s).c4);      

      ud+=1;
      pl+=1;
   }
}


void apply_Udag2sd(int vol,int *imb,su3_dble *ud,spinor_dble *pk,
                   spinor_dble *pl)
{
   int *imm;
   spinor_dble *s;

   imm=imb+vol;

   for (;imb<imm;imb++)
   {
      s=pk+(*imb);

      _su3_inverse_multiply((*pl).c1,(*ud),(*s).c1);
      _su3_inverse_multiply((*pl).c2,(*ud),(*s).c2);
      _su3_inverse_multiply((*pl).c3,(*ud),(*s).c3);
      _su3_inverse_multiply((*pl).c4,(*ud),(*s).c4);      

      ud+=1;
      pl+=1;
   }
}  

#endif

static void init_sm(void)
{
   int n;

   for (n=0;n<MAX_LEVELS;n++)
   {
      cnt[n]=0;
      sm0[n].re=0.0;
      sm0[n].im=0.0;
      sm1[n].re=0.0;
      sm1[n].im=0.0;
   }   
}


static void acc_sm(void)
{
   int n;
   
   cnt[0]+=1;
   sm0[0].re+=z0.re;
   sm0[0].im+=z0.im;
   sm1[0].re+=z1.re;
   sm1[0].im+=z1.im;
   
   for (n=1;(cnt[n-1]>=BLK_LENGTH)&&(n<MAX_LEVELS);n++)
   {
      cnt[n]+=1;
      sm0[n].re+=sm0[n-1].re;
      sm0[n].im+=sm0[n-1].im;
      sm1[n].re+=sm1[n-1].re;
      sm1[n].im+=sm1[n-1].im;      

      cnt[n-1]=0;
      sm0[n-1].re=0.0;
      sm0[n-1].im=0.0;
      sm1[n-1].re=0.0;
      sm1[n-1].im=0.0;
   }
}


static void sum_sm(void)
{
   int n;

   z0.re=0.0;
   z0.im=0.0;
   z1.re=0.0;
   z1.im=0.0;

   for (n=0;n<MAX_LEVELS;n++)
   {
      z0.re+=sm0[n].re;
      z0.im+=sm0[n].im;      
      z1.re+=sm1[n].re;
      z1.im+=sm1[n].im;      
   }
}

#if (defined SSE2)

#define _zgen_init_sp() \
__asm__ __volatile__ ("xorpd %%xmm6, %%xmm6 \n\t" \
                      "xorpd %%xmm7, %%xmm7" \
                      : \
                      : \
                      : \
                      "xmm6", "xmm7")

#define _zgen_store_sp() \
__asm__ __volatile__ ("movapd %%xmm6, %0 \n\t" \
                      "movapd %%xmm7, %1" \
                      : \
                      "=m" (z0.re), \
                      "=m" (z1.re))


#if (defined SSE3)

#define _zgen_load_chi(s) \
__asm__ __volatile__ ("movddup %0, %%xmm0 \n\t" \
                      "movddup %1, %%xmm1 \n\t" \
                      "movddup %2, %%xmm2 \n\t" \
                      "movddup %3, %%xmm3 \n\t" \
                      "movddup %4, %%xmm4 \n\t" \
                      "movddup %5, %%xmm5" \
                      : \
                      : \
                      "m" ((s).c1.re), \
                      "m" ((s).c1.im), \
                      "m" ((s).c2.re), \
                      "m" ((s).c2.im), \
                      "m" ((s).c3.re), \
                      "m" ((s).c3.im) \
                      : \
                      "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5")

#else

#define _zgen_load_chi(s) \
__asm__ __volatile__ ("movsd %0, %%xmm0 \n\t" \
                      "movsd %1, %%xmm1 \n\t" \
                      "movsd %2, %%xmm2 \n\t" \
                      "movsd %3, %%xmm3 \n\t" \
                      "movsd %4, %%xmm4 \n\t" \
                      "movsd %5, %%xmm5 \n\t" \
                      "unpcklpd %%xmm0, %%xmm0 \n\t" \
                      "unpcklpd %%xmm1, %%xmm1 \n\t" \
                      "unpcklpd %%xmm2, %%xmm2 \n\t" \
                      "unpcklpd %%xmm3, %%xmm3 \n\t" \
                      "unpcklpd %%xmm4, %%xmm4 \n\t" \
                      "unpcklpd %%xmm5, %%xmm5" \
                      : \
                      : \
                      "m" ((s).c1.re), \
                      "m" ((s).c1.im), \
                      "m" ((s).c2.re), \
                      "m" ((s).c2.im), \
                      "m" ((s).c3.re), \
                      "m" ((s).c3.im) \
                      : \
                      "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5")
   
#endif

#define _zgen_mulby_psi(s) \
__asm__ __volatile__ ("mulpd %0, %%xmm0 \n\t" \
                      "mulpd %0, %%xmm1 \n\t" \
                      "mulpd %1, %%xmm2 \n\t" \
                      "mulpd %1, %%xmm3 \n\t" \
                      "mulpd %2, %%xmm4 \n\t" \
                      "mulpd %2, %%xmm5 \n\t" \
                      : \
                      : \
                      "m" ((s).c1.re), \
                      "m" ((s).c2.re), \
                      "m" ((s).c3.re) \
                      : \
                      "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5")


#define _zgen_add_sp0() \
__asm__ __volatile__ ("addpd %%xmm0, %%xmm2 \n\t" \
                      "addpd %%xmm1, %%xmm3 \n\t" \
                      "addpd %%xmm2, %%xmm4 \n\t" \
                      "addpd %%xmm3, %%xmm5 \n\t" \
                      "mulpd %0, %%xmm4 \n\t" \
                      "shufpd $0x1, %%xmm5, %%xmm5 \n\t" \
                      "addpd %%xmm4, %%xmm6 \n\t" \
                      "addpd %%xmm5, %%xmm6" \
                      : \
                      : \
                      "m" (sgn) \
                      : \
                      "xmm2", "xmm3", "xmm4", "xmm5", "xmm6")


#define _zgen_add_sp1() \
__asm__ __volatile__ ("addpd %%xmm0, %%xmm2 \n\t" \
                      "addpd %%xmm1, %%xmm3 \n\t" \
                      "addpd %%xmm2, %%xmm4 \n\t" \
                      "addpd %%xmm3, %%xmm5 \n\t" \
                      "mulpd %0, %%xmm4 \n\t" \
                      "shufpd $0x1, %%xmm5, %%xmm5 \n\t" \
                      "addpd %%xmm4, %%xmm7 \n\t" \
                      "addpd %%xmm5, %%xmm7" \
                      : \
                      : \
                      "m" (sgn) \
                      : \
                      "xmm2", "xmm3", "xmm4", "xmm5", "xmm7")

#if (defined SSE3)
   
#define _zgen_sub_sp1() \
__asm__ __volatile__ ("addpd %%xmm0, %%xmm2 \n\t" \
                      "addpd %%xmm1, %%xmm3 \n\t" \
                      "addpd %%xmm2, %%xmm4 \n\t" \
                      "addpd %%xmm3, %%xmm5 \n\t" \
                      "addsubpd %%xmm4, %%xmm7 \n\t" \
                      "shufpd $0x1, %%xmm5, %%xmm5 \n\t" \
                      "subpd %%xmm5, %%xmm7" \
                      : \
                      : \
                      : \
                      "xmm2", "xmm3", "xmm4", "xmm5", "xmm7")

#else

#define _zgen_sub_sp1() \
__asm__ __volatile__ ("addpd %%xmm0, %%xmm2 \n\t" \
                      "addpd %%xmm1, %%xmm3 \n\t" \
                      "addpd %%xmm2, %%xmm4 \n\t" \
                      "addpd %%xmm3, %%xmm5 \n\t" \
                      "mulpd %0, %%xmm4 \n\t" \
                      "shufpd $0x1, %%xmm5, %%xmm5 \n\t" \
                      "subpd %%xmm4, %%xmm7 \n\t" \
                      "subpd %%xmm5, %%xmm7" \
                      : \
                      : \
                      "m" (sgn) \
                      : \
                      "xmm2", "xmm3", "xmm4", "xmm5", "xmm7")
   
#endif

   
static void spinor_prod_gamma0(int vol,spinor_dble *pk,spinor_dble *pl,
                               complex_dble *sp)
{
   spinor_dble *pt,*pm;

   init_sm();
   
   pt=pk+vol;
   pm=pk;

   for (;pm<pt;)
   {
      pm+=BLK_LENGTH;
      if (pm>pt)
         pm=pt;

      _zgen_init_sp();      

      for (;pk<pm;)
      {
         _zgen_load_chi( (*pl).c1);         
         _zgen_mulby_psi((*pk).c1);
         _zgen_add_sp0();
         _zgen_load_chi( (*pl).c2);         
         _zgen_mulby_psi((*pk).c2);
         _zgen_add_sp0();
         _zgen_load_chi( (*pl).c3);         
         _zgen_mulby_psi((*pk).c3);
         _zgen_add_sp0();
         _zgen_load_chi( (*pl).c4);         
         _zgen_mulby_psi((*pk).c4);
         _zgen_add_sp0();

         _zgen_load_chi( (*pl).c3);         
         _zgen_mulby_psi((*pk).c1);
         _zgen_add_sp1();
         _zgen_load_chi( (*pl).c4);         
         _zgen_mulby_psi((*pk).c2);
         _zgen_add_sp1();
         _zgen_load_chi( (*pl).c1);         
         _zgen_mulby_psi((*pk).c3);
         _zgen_add_sp1();
         _zgen_load_chi( (*pl).c2);         
         _zgen_mulby_psi((*pk).c4);
         _zgen_add_sp1();         

         pk+=1;
         pl+=1;
      }

      _zgen_store_sp();
      acc_sm();
   }

   sum_sm();
   
   sp[0].re=z0.re;
   sp[0].im=z0.im;
   
   sp[1].re=-z1.re;
   sp[1].im=-z1.im;
}


static void spinor_prod_gamma1(int vol,spinor_dble *pk,spinor_dble *pl,
                               complex_dble *sp)
{
   spinor_dble *pt,*pm;

   init_sm();
   
   pt=pk+vol;
   pm=pk;

   for (;pm<pt;)
   {
      pm+=BLK_LENGTH;
      if (pm>pt)
         pm=pt;

      _zgen_init_sp();      

      for (;pk<pm;)
      {
         _zgen_load_chi( (*pl).c1);         
         _zgen_mulby_psi((*pk).c1);
         _zgen_add_sp0();
         _zgen_load_chi( (*pl).c2);         
         _zgen_mulby_psi((*pk).c2);
         _zgen_add_sp0();
         _zgen_load_chi( (*pl).c3);         
         _zgen_mulby_psi((*pk).c3);
         _zgen_add_sp0();
         _zgen_load_chi( (*pl).c4);         
         _zgen_mulby_psi((*pk).c4);
         _zgen_add_sp0();

         _zgen_load_chi( (*pl).c4);         
         _zgen_mulby_psi((*pk).c1);
         _zgen_add_sp1();
         _zgen_load_chi( (*pl).c3);         
         _zgen_mulby_psi((*pk).c2);
         _zgen_add_sp1();
         _zgen_load_chi( (*pl).c2);         
         _zgen_mulby_psi((*pk).c3);
         _zgen_sub_sp1();
         _zgen_load_chi( (*pl).c1);         
         _zgen_mulby_psi((*pk).c4);
         _zgen_sub_sp1();         

         pk+=1;
         pl+=1;
      }

      _zgen_store_sp();
      acc_sm();
   }

   sum_sm();
   
   sp[0].re=z0.re;
   sp[0].im=z0.im;
   
   sp[1].re=z1.im;
   sp[1].im=-z1.re;
}


static void spinor_prod_gamma2(int vol,spinor_dble *pk,spinor_dble *pl,
                               complex_dble *sp)
{
   spinor_dble *pt,*pm;

   init_sm();
   
   pt=pk+vol;
   pm=pk;

   for (;pm<pt;)
   {
      pm+=BLK_LENGTH;
      if (pm>pt)
         pm=pt;

      _zgen_init_sp();      

      for (;pk<pm;)
      {
         _zgen_load_chi( (*pl).c1);         
         _zgen_mulby_psi((*pk).c1);
         _zgen_add_sp0();
         _zgen_load_chi( (*pl).c2);         
         _zgen_mulby_psi((*pk).c2);
         _zgen_add_sp0();
         _zgen_load_chi( (*pl).c3);         
         _zgen_mulby_psi((*pk).c3);
         _zgen_add_sp0();
         _zgen_load_chi( (*pl).c4);         
         _zgen_mulby_psi((*pk).c4);
         _zgen_add_sp0();

         _zgen_load_chi( (*pl).c4);         
         _zgen_mulby_psi((*pk).c1);
         _zgen_add_sp1();
         _zgen_load_chi( (*pl).c3);         
         _zgen_mulby_psi((*pk).c2);
         _zgen_sub_sp1();
         _zgen_load_chi( (*pl).c2);         
         _zgen_mulby_psi((*pk).c3);
         _zgen_sub_sp1();
         _zgen_load_chi( (*pl).c1);         
         _zgen_mulby_psi((*pk).c4);
         _zgen_add_sp1();         

         pk+=1;
         pl+=1;
      }

      _zgen_store_sp();
      acc_sm();
   }

   sum_sm();
   
   sp[0].re=z0.re;
   sp[0].im=z0.im;
   
   sp[1].re=-z1.re;
   sp[1].im=-z1.im;
}


static void spinor_prod_gamma3(int vol,spinor_dble *pk,spinor_dble *pl,
                               complex_dble *sp)
{
   spinor_dble *pt,*pm;

   init_sm();
   
   pt=pk+vol;
   pm=pk;

   for (;pm<pt;)
   {
      pm+=BLK_LENGTH;
      if (pm>pt)
         pm=pt;

      _zgen_init_sp();      

      for (;pk<pm;)
      {
         _zgen_load_chi( (*pl).c1);         
         _zgen_mulby_psi((*pk).c1);
         _zgen_add_sp0();
         _zgen_load_chi( (*pl).c2);         
         _zgen_mulby_psi((*pk).c2);
         _zgen_add_sp0();
         _zgen_load_chi( (*pl).c3);         
         _zgen_mulby_psi((*pk).c3);
         _zgen_add_sp0();
         _zgen_load_chi( (*pl).c4);         
         _zgen_mulby_psi((*pk).c4);
         _zgen_add_sp0();

         _zgen_load_chi( (*pl).c3);         
         _zgen_mulby_psi((*pk).c1);
         _zgen_add_sp1();
         _zgen_load_chi( (*pl).c4);         
         _zgen_mulby_psi((*pk).c2);
         _zgen_sub_sp1();
         _zgen_load_chi( (*pl).c1);         
         _zgen_mulby_psi((*pk).c3);
         _zgen_sub_sp1();
         _zgen_load_chi( (*pl).c2);         
         _zgen_mulby_psi((*pk).c4);
         _zgen_add_sp1();         

         pk+=1;
         pl+=1;
      }

      _zgen_store_sp();
      acc_sm();
   }

   sum_sm();
   
   sp[0].re=z0.re;
   sp[0].im=z0.im;
   
   sp[1].re=z1.im;
   sp[1].im=-z1.re;
}



#else

static void spinor_prod_gamma0(int vol,spinor_dble *pk,spinor_dble *pl,
                               complex_dble *sp)
{
   spinor_dble *pt,*pm;

   init_sm();
   
   pt=pk+vol;
   pm=pk;

   for (;pm<pt;)
   {
      pm+=BLK_LENGTH;
      if (pm>pt)
         pm=pt;

      z0.re=0.0;
      z0.im=0.0;
      z1.re=0.0;
      z1.im=0.0;      
      
      for (;pk<pm;)
      {
         z0.re+=(_vector_prod_re((*pk).c1,(*pl).c1)+
                 _vector_prod_re((*pk).c2,(*pl).c2)+
                 _vector_prod_re((*pk).c3,(*pl).c3)+
                 _vector_prod_re((*pk).c4,(*pl).c4));

         z0.im+=(_vector_prod_im((*pk).c1,(*pl).c1)+
                 _vector_prod_im((*pk).c2,(*pl).c2)+
                 _vector_prod_im((*pk).c3,(*pl).c3)+
                 _vector_prod_im((*pk).c4,(*pl).c4));      

         z1.re+=(_vector_prod_re((*pk).c1,(*pl).c3)+
                 _vector_prod_re((*pk).c2,(*pl).c4)+
                 _vector_prod_re((*pk).c3,(*pl).c1)+
                 _vector_prod_re((*pk).c4,(*pl).c2));               

         z1.im+=(_vector_prod_im((*pk).c1,(*pl).c3)+
                 _vector_prod_im((*pk).c2,(*pl).c4)+
                 _vector_prod_im((*pk).c3,(*pl).c1)+
                 _vector_prod_im((*pk).c4,(*pl).c2));               

         pk+=1;
         pl+=1;
      }

      acc_sm();
   }

   sum_sm();
   
   sp[0].re=z0.re;
   sp[0].im=z0.im;
   
   sp[1].re=-z1.re;
   sp[1].im=-z1.im;
}


static void spinor_prod_gamma1(int vol,spinor_dble *pk,spinor_dble *pl,
                               complex_dble *sp)
{
   spinor_dble *pt,*pm;

   init_sm();
   
   pt=pk+vol;
   pm=pk;

   for (;pm<pt;)
   {
      pm+=BLK_LENGTH;
      if (pm>pt)
         pm=pt;

      z0.re=0.0;
      z0.im=0.0;
      z1.re=0.0;
      z1.im=0.0;      
      
      for (;pk<pm;)
      {
         z0.re+=(_vector_prod_re((*pk).c1,(*pl).c1)+
                 _vector_prod_re((*pk).c2,(*pl).c2)+
                 _vector_prod_re((*pk).c3,(*pl).c3)+
                 _vector_prod_re((*pk).c4,(*pl).c4));

         z0.im+=(_vector_prod_im((*pk).c1,(*pl).c1)+
                 _vector_prod_im((*pk).c2,(*pl).c2)+
                 _vector_prod_im((*pk).c3,(*pl).c3)+
                 _vector_prod_im((*pk).c4,(*pl).c4));      

         z1.re+=(_vector_prod_re((*pk).c1,(*pl).c4)+
                 _vector_prod_re((*pk).c2,(*pl).c3));

         z1.re-=(_vector_prod_re((*pk).c3,(*pl).c2)+
                 _vector_prod_re((*pk).c4,(*pl).c1));               

         z1.im+=(_vector_prod_im((*pk).c1,(*pl).c4)+
                 _vector_prod_im((*pk).c2,(*pl).c3));

         z1.im-=(_vector_prod_im((*pk).c3,(*pl).c2)+
                 _vector_prod_im((*pk).c4,(*pl).c1));               

         pk+=1;
         pl+=1;
      }

      acc_sm();
   }

   sum_sm();
   
   sp[0].re=z0.re;
   sp[0].im=z0.im;
   
   sp[1].re=z1.im;
   sp[1].im=-z1.re;
}


static void spinor_prod_gamma2(int vol,spinor_dble *pk,spinor_dble *pl,
                               complex_dble *sp)
{
   spinor_dble *pt,*pm;

   init_sm();
   
   pt=pk+vol;
   pm=pk;

   for (;pm<pt;)
   {
      pm+=BLK_LENGTH;
      if (pm>pt)
         pm=pt;

      z0.re=0.0;
      z0.im=0.0;
      z1.re=0.0;
      z1.im=0.0;      
      
      for (;pk<pm;)
      {
         z0.re+=(_vector_prod_re((*pk).c1,(*pl).c1)+
                 _vector_prod_re((*pk).c2,(*pl).c2)+
                 _vector_prod_re((*pk).c3,(*pl).c3)+
                 _vector_prod_re((*pk).c4,(*pl).c4));

         z0.im+=(_vector_prod_im((*pk).c1,(*pl).c1)+
                 _vector_prod_im((*pk).c2,(*pl).c2)+
                 _vector_prod_im((*pk).c3,(*pl).c3)+
                 _vector_prod_im((*pk).c4,(*pl).c4));      

         z1.re+=(_vector_prod_re((*pk).c1,(*pl).c4)+
                 _vector_prod_re((*pk).c4,(*pl).c1));

         z1.re-=(_vector_prod_re((*pk).c2,(*pl).c3)+
                 _vector_prod_re((*pk).c3,(*pl).c2));               

         z1.im+=(_vector_prod_im((*pk).c1,(*pl).c4)+
                 _vector_prod_im((*pk).c4,(*pl).c1));

         z1.im-=(_vector_prod_im((*pk).c2,(*pl).c3)+
                 _vector_prod_im((*pk).c3,(*pl).c2));               

         pk+=1;
         pl+=1;
      }

      acc_sm();
   }

   sum_sm();
   
   sp[0].re=z0.re;
   sp[0].im=z0.im;
   
   sp[1].re=-z1.re;
   sp[1].im=-z1.im;
}


static void spinor_prod_gamma3(int vol,spinor_dble *pk,spinor_dble *pl,
                               complex_dble *sp)
{
   spinor_dble *pt,*pm;

   init_sm();
   
   pt=pk+vol;
   pm=pk;

   for (;pm<pt;)
   {
      pm+=BLK_LENGTH;
      if (pm>pt)
         pm=pt;

      z0.re=0.0;
      z0.im=0.0;
      z1.re=0.0;
      z1.im=0.0;      
      
      for (;pk<pm;)
      {
         z0.re+=(_vector_prod_re((*pk).c1,(*pl).c1)+
                 _vector_prod_re((*pk).c2,(*pl).c2)+
                 _vector_prod_re((*pk).c3,(*pl).c3)+
                 _vector_prod_re((*pk).c4,(*pl).c4));

         z0.im+=(_vector_prod_im((*pk).c1,(*pl).c1)+
                 _vector_prod_im((*pk).c2,(*pl).c2)+
                 _vector_prod_im((*pk).c3,(*pl).c3)+
                 _vector_prod_im((*pk).c4,(*pl).c4));      

         z1.re+=(_vector_prod_re((*pk).c1,(*pl).c3)+
                 _vector_prod_re((*pk).c4,(*pl).c2));

         z1.re-=(_vector_prod_re((*pk).c2,(*pl).c4)+
                 _vector_prod_re((*pk).c3,(*pl).c1));               

         z1.im+=(_vector_prod_im((*pk).c1,(*pl).c3)+
                 _vector_prod_im((*pk).c4,(*pl).c2));

         z1.im-=(_vector_prod_im((*pk).c2,(*pl).c4)+
                 _vector_prod_im((*pk).c3,(*pl).c1));               

         pk+=1;
         pl+=1;
      }

      acc_sm();
   }

   sum_sm();
   
   sp[0].re=z0.re;
   sp[0].im=z0.im;
   
   sp[1].re=z1.im;
   sp[1].im=-z1.re;
}

#endif

void (*spinor_prod_gamma[4])
   (int vol,spinor_dble *pk,spinor_dble *pl,complex_dble *sp)=
{spinor_prod_gamma0,spinor_prod_gamma1,spinor_prod_gamma2,spinor_prod_gamma3};

