
/*******************************************************************************
*
* File Zhat_dble.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Application of the double-precision little Dirac operator
*
* The externally accessible functions are
*
*   void Zhat_dble(complex_dble *v,complex_dble *w)
*     Applies the even-odd preconditioned little Dirac operator to the
*     even components of the global double-precision vector field v[] and
*     assigns the result to the field w[]
*
*   void Zeeinv_dble(complex_dble *v,complex_dble *w)
*     Applies the inverse of the even-even part of the little Dirac operator
*     to the global double-precision vector field v[] and assigns the result
*     to the field w[]
*
*   void Zooinv_dble(complex_dble *v,complex_dble *w)
*     Applies the inverse of the odd-odd part of the little Dirac operator
*     to the global double-precision vector field v[] and assigns the result
*     to the field w[]
*
*   void Zoe_dble(complex_dble *v,complex_dble *w)
*     Applies the odd-even part of the little Dirac operator to the global
*     double-precision vector field v[] and assigns the result to the
*     field w[]
*
*   void Zeo_dble(complex_dble *v,complex_dble *w)
*     Applies the even-odd part of the little Dirac operator to the global
*     double-precision vector field v[] and *subtracts* the result from
*     the field w[]
*
*   void Znohat_dble(complex_dble *v,complex_dble *w)
*     Applies the little Dirac operator to the global double-precision
*     vector field v[] and assigns the result to the field w[]
*
* Notes:
*
* The restriction of the (improved) Wilson-Dirac operator Dw to the deflation
* subspace is referred to as the little Dirac operator. Its matrix elements
* (and those of the even-odd preconditioned operator) are computed by the
* appropriate programs in the module dfl/Zops.c
*
* The global vector fields are assumed to be those allocated by the programs 
* in the module dfl/vflds.c. Input and output fields must be different (the
* behaviour is otherwise undefined)
* 
* All programs in this module perform global operations and should therefore
* be called simultaneously on all processes
*
*******************************************************************************/

#define ZHAT_DBLE_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "misc.h"
#include "dfl.h"
#include "global.h"

static int Nss=0;
static complex_dble *rds;


static void alloc_rds(int Ns)
{
   if (Nss>0)
      afree(rds);

   rds=amalloc(Ns*sizeof(*rds),ALIGN);
   
   error(rds==NULL,1,"alloc_rds [Zhat_dble.c]",
         "Unable to allocate auxiliary array");

   Nss=Ns;
}


static void apply_zoe_dble(int Ns,int *in,complex_dble *z,complex_dble *v)
{
   int nmat,*im;
   
   cmat_vec_dble(Ns,z,v+(*in)*Ns,rds);

   nmat=Ns*Ns;   
   z+=nmat;
   im=in+8;
   in+=1;

   for (;in<im;in++)
   {
      cmat_vec_assign_dble(Ns,z,v+(*in)*Ns,rds);
      z+=nmat;
   }
}


static void apply_zeo_dble(int Ns,int *in,complex_dble *z,complex_dble *v)
{
   int nmat,*im;
   
   nmat=Ns*Ns;   
   im=in+8;

   for (;in<im;in++)
   {
      cmat_vec_assign_dble(Ns,z,rds,v+(*in)*Ns);
      z+=nmat;
   }
}


void Zhat_dble(complex_dble *v,complex_dble *w)
{
   int nb,nbb,Ns,nz,n;
   int (*inn)[8],*in,*im;
   complex_dble *zoe,*zeo,*r,*rm;
   zop_dble_t *zop;

   zop=zophat_dble();
   nb=(*zop).nb;
   nbb=(*zop).nbb;
   Ns=(*zop).Ns;
   nz=8*Ns*Ns;

   if (Ns>Nss)
      alloc_rds(Ns);
   
   assign_vd2vd((nb/2)*Ns,v,w);
   
   if (NPROC>1)
   {
      set_vd2zero((nbb/2)*Ns,w+nb*Ns);
      cpvd_int_bnd(v);
   }
   
   inn=vinn(&n);
   in=inn[nb/2];
   im=in+nb*4;
   zoe=(*zop).zoe;
   zeo=(*zop).zeo;

   for (;in<im;in+=8)
   {
      apply_zoe_dble(Ns,in,zoe,v);

      r=rds;
      rm=r+Ns;

      for (;r<rm;r++)
      {
         (*r).re=-(*r).re;
         (*r).im=-(*r).im;
      }

      apply_zeo_dble(Ns,in,zeo,w);      

      zoe+=nz;
      zeo+=nz;
   }

   if (NPROC>1)
      cpvd_ext_bnd(w);   
}


static void apply_zdiag_dble(int n,int Ns,complex_dble *z,complex_dble *v,
                             complex_dble *w)
{
   int nmat;
   complex_dble *vm;

   nmat=Ns*Ns;
   vm=v+n*Ns;

   for (;v<vm;v+=Ns)
   {
      cmat_vec_dble(Ns,z,v,w);      
      z+=nmat;
      w+=Ns;
   }
}


void Zeeinv_dble(complex_dble *v,complex_dble *w)
{
   int nb,Ns;
   zop_dble_t *zop;

   zop=zophat_dble();
   nb=(*zop).nb;
   Ns=(*zop).Ns;
   
   apply_zdiag_dble(nb/2,Ns,(*zop).zee,v,w);
}


void Zooinv_dble(complex_dble *v,complex_dble *w)
{
   int nb,Ns,n;
   zop_dble_t *zop;

   zop=zophat_dble();
   nb=(*zop).nb;
   Ns=(*zop).Ns;
   n=(nb/2)*Ns;
   
   apply_zdiag_dble(nb/2,Ns,(*zop).zoo,v+n,w+n);
}


void Zoe_dble(complex_dble *v,complex_dble *w)
{
   int nb,Ns,nz,n;
   int (*inn)[8],*in,*im;
   complex_dble *zoe,*r,*rm,*rw;
   zop_dble_t *zop;

   zop=zopnohat_dble();
   nb=(*zop).nb;
   Ns=(*zop).Ns;
   nz=8*Ns*Ns;

   if (Ns>Nss)
      alloc_rds(Ns);
   
   if (NPROC>1)
      cpvd_int_bnd(v);
   
   inn=vinn(&n);
   in=inn[nb/2];
   im=in+nb*4;
   zoe=(*zop).zoe;
   rw=w+(nb/2)*Ns;

   for (;in<im;in+=8)
   {
      apply_zoe_dble(Ns,in,zoe,v);

      r=rds;
      rm=r+Ns;

      for (;r<rm;r++)
      {
         (*rw).re=(*r).re;
         (*rw).im=(*r).im;
         rw+=1;
      }

      zoe+=nz;
   }
}


void Zeo_dble(complex_dble *v,complex_dble *w)
{
   int nb,nbb,Ns,nz,n;
   int (*inn)[8],*in,*im;
   complex_dble *zeo,*r,*rm,*rv;
   zop_dble_t *zop;

   zop=zopnohat_dble();
   nb=(*zop).nb;
   nbb=(*zop).nbb;
   Ns=(*zop).Ns;
   nz=8*Ns*Ns;

   if (Ns>Nss)
      alloc_rds(Ns);
   
   if (NPROC>1)
      set_vd2zero((nbb/2)*Ns,w+nb*Ns);
   
   inn=vinn(&n);
   in=inn[nb/2];
   im=in+nb*4;
   zeo=(*zop).zeo;
   rv=v+(nb/2)*Ns;

   for (;in<im;in+=8)
   {
      r=rds;
      rm=r+Ns;

      for (;r<rm;r++)
      {
         (*r).re=-(*rv).re;
         (*r).im=-(*rv).im;
         rv+=1;
      }
      
      apply_zeo_dble(Ns,in,zeo,w);

      zeo+=nz;
   }

   if (NPROC>1)
      cpvd_ext_bnd(w);    
}


void Znohat_dble(complex_dble *v,complex_dble *w)
{
   int nb,nbb,Ns,nz,n;
   int (*inn)[8],*in,*im;
   complex_dble *zoe,*zeo,*r,*rm,*rv,*rw;
   zop_dble_t *zop;

   zop=zopnohat_dble();
   nb=(*zop).nb;
   nbb=(*zop).nbb;
   Ns=(*zop).Ns;
   nz=8*Ns*Ns;

   if (Ns>Nss)
      alloc_rds(Ns);
   
   apply_zdiag_dble(nb,Ns,(*zop).zee,v,w);
   
   if (NPROC>1)
   {
      set_vd2zero((nbb/2)*Ns,w+nb*Ns);      
      cpvd_int_bnd(v);
   }
   
   inn=vinn(&n);
   in=inn[nb/2];
   im=in+nb*4;
   zoe=(*zop).zoe;
   zeo=(*zop).zeo;
   rv=v+(nb/2)*Ns;
   rw=w+(nb/2)*Ns;

   for (;in<im;in+=8)
   {
      apply_zoe_dble(Ns,in,zoe,v);

      r=rds;
      rm=r+Ns;

      for (;r<rm;r++)
      {
         (*rw).re+=(*r).re;
         (*rw).im+=(*r).im;

         (*r).re=(*rv).re;
         (*r).im=(*rv).im;
         
         rw+=1;
         rv+=1;
      }
      
      apply_zeo_dble(Ns,in,zeo,w);      

      zoe+=nz;
      zeo+=nz;
   }

   if (NPROC>1)
      cpvd_ext_bnd(w);   
}

