
/*******************************************************************************
*
* File dfl.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Application of the deflation preconditioner
*
* The externally accessible functions are
*
*   void dfl_Pproj_dble(int k,int l)
*     Applies the projector to the deflation subspace to the global
*     double-precision spinor field psd[k] and assigns the result to 
*     psd[l] (where l may be equal to k)
*
*   void dfl_Qproj_dble(int k,int l)
*     Applies the projector to the orthogonal complement of the deflation
*     subspace to the global double-precision spinor field psd[k] and 
*     assigns the result to psd[l] (where l may be equal to k)
*
*   double dfl_solve_little(int nmx,double res,complex_dble *v,int *status)
*     Obtains an approximate solution of the little Dirac equation for
*     a given source field v, with a relative residue smaller than res and
*     using at most nmx deflated even-odd preconditioned GCR iterations.
*     On exit the solution is assigned to v and status reports the number
*     of iterations that were required (or a negative number if the solver
*     failed). The return value is the norm of the residue of the solution
*     
*   void dfl_Lpcnd_dble(int k,int l,int *status)
*     Applies the deflation left-preconditioner to the global double-
*     precision spinor field psd[k] and solves the Dirac equation taking
*     the complement of the preconditioned field as source. On exit the
*     preconditioned field and the solution of the Dirac equation are
*     assigned to the fields psd[k] and psd[l], while status reports the
*     number of solver iterations that were required for the solution of
*     the little Dirac equation (or a negative number if the solver failed)
*
*   void dfl_RLpcnd_dble(int k,int l,int *status)
*     Applies the deflation right- and left-preconditioners to the global
*     double-precision spinor fields psd[k] and psd[l], assuming they are 
*     related through psd[l]=Dw*psd[k], Dw being the Wilson-Dirac operator.
*     On exit status reports the number of solver iterations that were
*     required for the solution of the little Dirac equation (or a negative
*     number if the solver failed)
*
*   void dfl_RLpcnd(int k,int l,int *status)
*     Applies the deflation right- and left-preconditioners to the global
*     single-precision spinor fields ps[k] and ps[l], assuming they are 
*     related through ps[l]=Dw*ps[k], Dw being the Wilson-Dirac operator.
*     On exit status reports the number of solver iterations that were
*     required for the solution of the little Dirac equation (or a negative
*     number if the solver failed)
*
* Notes:
*
* The programs in this module take it for granted that the parameters of
* the deflation preconditioner have been set by the program set_dfl_parms()
* [flags/parms.c]. Moreover it is assumed that the deflation subspace was
* defined using the programs dfl_subspace() or dfl_modes()
*
* The solver for the little Dirac equation and the preconditioner programs
* need a number of single- and double-precision vector fields as workspace
* which are allocated when the parameters of the deflation preconditioner
* are set. In addition dfl_Lpcnd_dble() reserves 1 double-precision spinor
* field as workspace and dfl_RLpcnd() assumes the global single-precision
* gauge and SW fields to be allocated 
*
* The single-precision gauge field, the required SW arrays and the little
* Dirac operator are updated automatically when needed. No preparations 
* are thus required except for those mentioned above
*
* All programs in this module perform global operations and must be called
* simultaneously on all processes
*
*******************************************************************************/

#define DFL_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "flags.h"
#include "misc.h"
#include "linalg.h"
#include "linsolv.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "dfl.h"
#include "global.h"

static int nms=0,nvhs;
static complex **vs,**zvs;
static complex_dble **vds,**zvds,*avd,*cs1,*cs2;


static void alloc_arrays(int nm)
{
   if (nms>0)
   {
      afree(vs);
      afree(vds);
      afree(avd);
   }

   if (nm>0)
   {
      vs=amalloc(2*nm*sizeof(*vs),3);
      vds=amalloc(2*nm*sizeof(*vds),3);
      avd=amalloc((nm*nm+2*nm)*sizeof(*avd),ALIGN);   
      error((vs==NULL)||(vds==NULL)||(avd==NULL),1,
            "alloc_arrays [dfl.c]","Unable to allocate auxiliary arrays");

      zvs=vs+nm;
      zvds=vds+nm;
      cs1=avd+nm*nm;
      cs2=cs1+nm;
   }
   else
   {
      vs=NULL;
      zvs=NULL;
      vds=NULL;
      zvds=NULL;
      avd=NULL;
      cs1=NULL;
      cs2=NULL;
   }

   nms=nm;
}


void dfl_Pproj_dble(int k,int l)
{
   int iprms[2],nb,isw,nm,vol,i;
   spinor_dble **sd;
   block_t *b,*bm;
   dfl_parms_t dfl;

   if (NPROC>1)
   {
      iprms[0]=k;
      iprms[1]=l;

      MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=k)||(iprms[1]!=l),1,
            "dfl_Pproj_dble [dfl.c]","Parameters k,l are not global");
   }
   
   error_root((k<0)||(k>=no_sd)||(l<0)||(l>=no_sd),1,
              "dfl_Pproj_dble [dfl.c]","Improper arguments");
   
   dfl=dfl_parms();
   nm=dfl.Ns;

   if (nm==0)
   {
      set_sd2zero(VOLUME,psd[l][0]);
      return;
   }
   
   if (nm!=nms)
      alloc_arrays(nm);

   b=blk_list(DFL_BLOCKS,&nb,&isw);
   bm=b+nb;
   vol=(*b).vol;
   
   for (;b<bm;b++)
   {
      assign_sd2sdb(b,ALL_PTS,k,0);
      sd=(*b).sd;
      
      for (i=1;i<=nm;i++)
         cs1[i-1]=spinor_prod_dble(vol,0,sd[i],sd[0]);

      set_sd2zero(vol,sd[0]);

      for (i=1;i<=nm;i++)
         mulc_spinor_add_dble(vol,sd[0],sd[i],cs1[i-1]);
         
      assign_sdb2sd(b,ALL_PTS,0,l);
   }
}


void dfl_Qproj_dble(int k,int l)
{
   int iprms[2],nb,isw,nm,vol,i;
   spinor_dble **sd;
   block_t *b,*bm;
   dfl_parms_t dfl;

   if (NPROC>1)
   {
      iprms[0]=k;
      iprms[1]=l;

      MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=k)||(iprms[1]!=l),1,
            "dfl_Qproj_dble [dfl.c]","Parameters k,l are not global");
   }
   
   error_root((k<0)||(k>=no_sd)||(l<0)||(l>=no_sd),1,
              "dfl_Qproj_dble [dfl.c]","Improper arguments");   

   dfl=dfl_parms();
   nm=dfl.Ns;

   if (nm==0)
   {
      if (k!=l)
         assign_sd2sd(VOLUME,psd[k][0],psd[l][0]);
      return;
   }
   
   b=blk_list(DFL_BLOCKS,&nb,&isw);
   bm=b+nb;
   vol=(*b).vol;
   
   for (;b<bm;b++)
   {
      assign_sd2sdb(b,ALL_PTS,k,0);
      sd=(*b).sd;
      
      for (i=1;i<=nm;i++)
         project_dble(vol,0,sd[0],sd[i]);

      assign_sdb2sd(b,ALL_PTS,0,l);
   }
}


static void scale_vd(int n,complex_dble *v,double r)
{
   complex_dble *vm;

   vm=v+n;
   
   for (;v<vm;v++)
   {
      (*v).re*=r;
      (*v).im*=r;      
   }
}


static void sum_vprod(int n)
{
   int i;
   
   if (NPROC>1)
   {
      MPI_Reduce((double*)(cs1),(double*)(cs2),2*n,MPI_DOUBLE,
                 MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast((double*)(cs2),2*n,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
   else
   {
      for (i=0;i<n;i++)
      {
         cs2[i].re=cs1[i].re;
         cs2[i].im=cs1[i].im;
      }
   }
}


static void Lvd(complex_dble *v,complex_dble *w)
{
   int nm,nvh,i;
   complex_dble z;

   nm=nms;
   nvh=nvhs;

   for (i=0;i<nm;i++)
      cs1[i]=vprod_dble(nvh,0,vds[i],v);

   sum_vprod(nm);
   cmat_vec_dble(nm,avd,cs2,cs1);
   set_vd2zero(nvh,w);

   for (i=0;i<nm;i++)
   {
      mulc_vadd_dble(nvh,w,vds[i],cs1[i]);
      z.re=-cs1[i].re;
      z.im=-cs1[i].im;      
      mulc_vadd_dble(nvh,v,zvds[i],z);
   }
}


static void RLvd(complex_dble *v,complex_dble *w)
{
   int nm,nvh,i;
   complex_dble z;

   nm=nms;
   nvh=nvhs;

   for (i=0;i<nm;i++)
      cs1[i]=vprod_dble(nvh,0,vds[i],w);

   sum_vprod(nm);
   cmat_vec_dble(nm,avd,cs2,cs1);

   for (i=0;i<nm;i++)
   {
      z.re=-cs1[i].re;
      z.im=-cs1[i].im;      
      mulc_vadd_dble(nvh,v,vds[i],z);
      mulc_vadd_dble(nvh,w,zvds[i],z);
   }
}


static void Lv(complex *v)
{
   int nm,nvh,i;
   complex z;

   nm=nms;
   nvh=nvhs;

   for (i=0;i<nm;i++)
   {
      z=vprod(nvh,0,vs[i],v);
      cs1[i].re=(double)(z.re);
      cs1[i].im=(double)(z.im);      
   }

   sum_vprod(nm);
   cmat_vec_dble(nm,avd,cs2,cs1);

   for (i=0;i<nm;i++)
   {
      z.re=(float)(-cs1[i].re);
      z.im=(float)(-cs1[i].im);      
      mulc_vadd(nvh,v,zvs[i],z);
   }
}


static void Mop(int j,complex *pr,complex *pz,complex *paz)
{
   assign_v2v(nvhs,pr,pz);
   Zhat(pr,paz);
   Lv(paz);
}


static void Op(complex_dble *v,complex_dble *w)
{
   Zhat_dble(v,w);
   RLvd(v,w);
}


double dfl_solve_little(int nmx,double res,complex_dble *v,int *status)
{
   int nb,isw,nm,nkv,nvh;
   double r,rho;
   complex **wv;
   complex_dble **wvd,z;
   dfl_parms_t dfl;   

   dfl=dfl_parms();
   nm=dfl.Ns;
   nkv=dfl.vnkv;
   blk_list(DFL_BLOCKS,&nb,&isw);
   nvh=(nb/2)*nm;
   nvhs=nvh;

   if (nm!=nms)
      alloc_arrays(nm);

   dfl_little_modes(vs,zvs,vds,zvds,avd);
   wv=reserve_wv(2*nkv+1);
   wvd=reserve_wvd(3);

   Zooinv_dble(v,wvd[1]);
   Zeo_dble(wvd[1],v);
   Zeeinv_dble(v,wvd[1]);
   r=vnorm_square_dble(nvh,1,wvd[1]);

   if (r!=0.0)
   {
      r=sqrt(r);
      scale_vd(nvh,wvd[1],1.0/r);         
      Lvd(wvd[1],wvd[2]);
   
      rho=fgcr4vd(nvh,1,Op,Mop,wv,wvd,nkv,nmx,res,wvd[1],v,status);

      z.re=1.0;
      z.im=0.0;
      mulc_vadd_dble(nvh,v,wvd[2],z);
      scale_vd(nvh,v,r);
   }
   else
   {
      (*status)=0;
      rho=0.0;
      set_vd2zero(nvh,v);
   }

   Zoe_dble(v,wvd[1]);
   z.re=-1.0;
   z.im=0.0;
   mulc_vadd_dble(nvh,v+nvh,wvd[1]+nvh,z);
   assign_vd2vd(nvh,v+nvh,wvd[1]+nvh);
   Zooinv_dble(wvd[1],v);   

   release_wvd();   
   release_wv();

   return rho;
}


void dfl_Lpcnd_dble(int k,int l,int *status)
{
   int iprms[2],iw0;
   complex_dble **wvd;
   spinor_dble **wsd;
   dfl_parms_t dfl;

   if (NPROC>1)
   {
      iprms[0]=k;
      iprms[1]=l;      

      MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=k)||(iprms[1]!=l),1,"dfl_Lpcnd_dble [dfl.c]",
            "Parameters are not global");
   }
   
   error_root((k<0)||(k>=no_sd)||(l<0)||(l>=no_sd)||(k==l),1,
              "dfl_Lpcnd_dble [dfl.c]","Improper arguments");
   
   if ((query_flags(SWD_UP2DATE)!=1)||(query_flags(SWD_E_INVERTED)!=0)||
       (query_flags(SWD_O_INVERTED)!=0))
      sw_term();

   wvd=reserve_wvd(1);
   dfl_sd2vd(k,wvd[0]);
   dfl=dfl_parms();
   dfl_solve_little(dfl.vdnmx,dfl.vdres,wvd[0],status);
   dfl_vd2sd(wvd[0],l);
   release_wvd();
   
   wsd=reserve_wsd(1);
   iw0=(wsd[0]-psd[0][0])/NSPIN;  
   Qnohat_dble(l,iw0);
   mulg5_spinor_sub_dble(VOLUME,psd[k][0],wsd[0]);
   release_wsd();
}


void dfl_RLpcnd_dble(int k,int l,int *status)
{
   int iprms[2];
   complex_dble **wvd;
   dfl_parms_t dfl;
   
   if (NPROC>1)
   {
      iprms[0]=k;
      iprms[1]=l;

      MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=k)||(iprms[1]!=l),1,
            "dfl_RLpcnd_dble [dfl.c]","Parameters are not global");
   }
   
   error_root((k<0)||(k>=no_sd)||(l<0)||(l>=no_sd)||(k==l),1,
              "dfl_RLpcnd_dble [dfl.c]","Improper arguments");
   
   if ((query_flags(SWD_UP2DATE)!=1)||(query_flags(SWD_E_INVERTED)!=0)||
       (query_flags(SWD_O_INVERTED)!=0))
      sw_term();

   wvd=reserve_wvd(1);
   dfl_sd2vd(l,wvd[0]);
   dfl=dfl_parms();   
   dfl_solve_little(dfl.vdnmx,dfl.vdres,wvd[0],status);
   dfl_sub_vd2sd(wvd[0],k);
   release_wvd();
   
   Qnohat_dble(k,l);
   mulg5_dble(VOLUME,psd[l][0]);
}


void dfl_RLpcnd(int k,int l,int *status)
{
   int iprms[2],nb,isw,nv;
   complex **wv;
   complex_dble **wvd;
   dfl_parms_t dfl;

   if (NPROC>1)
   {
      iprms[0]=k;
      iprms[1]=l;

      MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=k)||(iprms[1]!=l),1,
            "dfl_RLpcnd [dfl.c]","Parameters are not global");
   }
   
   error_root((k<0)||(k>=no_s)||(l<0)||(l>=no_s)||(k==l),1,
              "dfl_RLpcnd [dfl.c]","Improper arguments");

   if (query_flags(U_MATCH_UD)!=1)
      assign_ud2u();
   
   if ((query_flags(SW_UP2DATE)!=1)||(query_flags(SW_E_INVERTED)!=0)||
       (query_flags(SW_O_INVERTED)!=0))
   {
      if ((query_flags(SWD_UP2DATE)!=1)||(query_flags(SWD_E_INVERTED)!=0)||
          (query_flags(SWD_O_INVERTED)!=0))
         sw_term();

      assign_swd2sw();
   }
   
   wv=reserve_wv(1);   
   dfl_s2v(l,wv[0]);

   wvd=reserve_wvd(1);   
   dfl=dfl_parms();
   blk_list(DFL_BLOCKS,&nb,&isw);
   nv=nb*dfl.Ns;
   assign_v2vd(nv,wv[0],wvd[0]);
   dfl_solve_little(dfl.vnmx,dfl.vres,wvd[0],status);
   assign_vd2v(nv,wvd[0],wv[0]);
   release_wvd();

   dfl_sub_v2s(wv[0],k);
   release_wv();

   Qnohat(k,l);
   mulg5(VOLUME,ps[l][0]);
}

