
/*******************************************************************************
*
* File dfl_subspace.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Basic utility programs related to the deflation subspace
*
* The externally accessible functions are
*
*   void dfl_sd2vd(int k,complex_dble *v)
*     Assigns the components of the global double-precision spinor field
*     psd[k] along the deflation subspace to the double-precision vector
*     field v
*
*   void dfl_vd2sd(complex_dble *v,int k)
*     Assigns the element of the deflation subspace corresponding to the
*     double-precision vector field v to the global double-precision spinor
*     field psd[k]
*
*   void dfl_sub_vd2sd(complex_dble *v,int k)
*     Subtracts the element of the deflation subspace corresponding to the
*     double-precision vector field v from the global double-precision spinor
*     field psd[k]
*
*   void dfl_s2v(int k,complex *v)
*     Assigns the components of the global single-precision spinor field
*     ps[k] along the deflation subspace to the single-precision vector
*     field v
*
*   void dfl_v2s(complex *v,int k)
*     Assigns the element of the deflation subspace corresponding to the
*     single-precision vector field v to the global single-precision spinor
*     field ps[k]
*
*   void dfl_sub_v2s(complex *v,int k)
*     Subtracts the element of the deflation subspace corresponding to the
*     double-precision vector field v from the global single-precision spinor
*     field ps[k]
*
*   void dfl_subspace(spinor **mds)
*     Copies the global single-precision spinor fields mds[0],..,mds[Ns-1] to 
*     the block fields (*b).sd[1],..,(*b).sd[Ns] on the DFL_BLOCKS grid and
*     orthonormalizes the latter using the modified Gram-Schmidt procedure.
*     The number Ns of fields to be copied is queried from the parameter data
*     base [flags/parms.c]
*
*   void dfl_global_modes(complex **v)
*     Sets v[0],..,v[Ns-1] to the pointers to Ns internally stored single-
*     precision vector fields that represent the global spinor fields last
*     imported to the DFL_BLOCKS grid by the program dfl_subspace() 
*
* Notes:
*
* The deflation subspace is spanned by the fields (*b).sd[1],..,(*b).sd[Ns]
* on the blocks b of the DFL_BLOCKS grid. The number Ns of fields is set by
* the program dfl_set_parms() [flags/parms.c] and it is taken for granted
* that this program is called before any programs in this module
*
* Any spinor field in the deflation subspace is a linear combination of the
* basis elements on the blocks. The associated complex coefficients form a
* vector field of the type described in dfl/vflds.c. Such vector fields are
* thus in one-to-one correspondence with the deflation modes
*
* The deflation subspace contains the global spinor fields from which the
* the subspace was created via the block projection. The corresponding
* vector fields are kept in the static memory and are made available through
* the program dfl_global_modes()
*
* When the deflation subspace is initialized, the basis block fields are 
* copied to the single-precision block fields  (*b).s[1],..,(*b).s[Ns] so
* that these are guaranteed to span the same subspace. Moreover, the little
* Dirac operator is marked as "unusable" in the flags data base. It will
* thus be automatically recomputed when needed
*
* All programs in this module perform global operations and must be called
* simultaneously on all processes
*
*******************************************************************************/

#define DFL_SUBSPACE_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "flags.h"
#include "linalg.h"
#include "block.h"
#include "dfl.h"
#include "global.h"

static int Nss=0,nvecs=0,*ism;
static complex **vs;


static void alloc_modes(void)
{
   int nb,isw,i;
   int nbl[4],Ns,nvec;
   complex *v;
   block_t *b;
   dfl_parms_t dfl;

   dfl=dfl_parms();
   Ns=dfl.Ns;
   b=blk_list(DFL_BLOCKS,&nb,&isw);

   error((Ns==0)||(b==NULL),1,"alloc_modes [dfl_subspace.c]",
         "The deflation parameters are not set or the DFL_BLOCKS\n"
         "grid is not allocated");
   
   nbl[0]=L0/(*b).bs[0];
   nbl[1]=L1/(*b).bs[1];
   nbl[2]=L2/(*b).bs[2];
   nbl[3]=L3/(*b).bs[3];

   nvec=nb;

   if (NPROC0>1)
      nvec+=nbl[1]*nbl[2]*nbl[3];
   if (NPROC1>1)
      nvec+=nbl[0]*nbl[2]*nbl[3];
   if (NPROC2>1)
      nvec+=nbl[0]*nbl[1]*nbl[3];
   if (NPROC3>1)
      nvec+=nbl[0]*nbl[1]*nbl[2];

   nvec*=Ns;

   if ((Nss==Ns)&&(nvecs==nvec))
      return;

   if (Nss>0)
   {
      afree(ism);
      afree(vs[0]);
      afree(vs);
   }

   ism=amalloc(Ns*sizeof(*ism),3);
   vs=amalloc(Ns*sizeof(*vs),3);
   v=amalloc(Ns*nvec*sizeof(*v),ALIGN);   
   
   error((ism==NULL)||(vs==NULL)||(v==NULL),1,
         "alloc_modes [dfl_subspace.c]","Unable to allocate vector fields");

   for (i=0;i<Ns;i++)
   {
      vs[i]=v;
      v+=nvec;
   }
   
   Nss=Ns;
   nvecs=nvec;
}


void dfl_sd2vd(int k,complex_dble *v)
{
   int nb,isw,Ns,i;
   complex_dble *vm;
   block_t *b0,*b,*bm;
   dfl_parms_t dfl;

   dfl=dfl_parms();
   Ns=dfl.Ns;
   b0=blk_list(DFL_BLOCKS,&nb,&isw);
   bm=b0+nb;
   if (isw==0)
      b=b0;
   else
      b=b0+(nb/2);
   vm=v+nb*Ns;

   for (;v<vm;)
   {
      assign_sd2sdb(b,ALL_PTS,k,0);
      
      for (i=1;i<=Ns;i++)
      {
         (*v)=spinor_prod_dble((*b).vol,0,(*b).sd[i],(*b).sd[0]);
         v+=1;
      }

      b+=1;
      if (b==bm)
         b=b0;
   }   
}


void dfl_vd2sd(complex_dble *v,int k)
{
   int nb,isw,Ns,i;
   complex_dble *vm;
   block_t *b0,*b,*bm;
   dfl_parms_t dfl;

   dfl=dfl_parms();
   Ns=dfl.Ns;
   b0=blk_list(DFL_BLOCKS,&nb,&isw);
   bm=b0+nb;
   if (isw==0)
      b=b0;
   else
      b=b0+(nb/2);
   vm=v+nb*Ns;

   for (;v<vm;)
   {
      set_sd2zero((*b).vol,(*b).sd[0]);
      
      for (i=1;i<=Ns;i++)
      {
         mulc_spinor_add_dble((*b).vol,(*b).sd[0],(*b).sd[i],*v);
         v+=1;
      }

      assign_sdb2sd(b,ALL_PTS,0,k);
      
      b+=1;
      if (b==bm)
         b=b0;
   }   
}


void dfl_sub_vd2sd(complex_dble *v,int k)
{
   int nb,isw,Ns,i;
   complex_dble *vm,z;
   block_t *b0,*b,*bm;
   dfl_parms_t dfl;

   dfl=dfl_parms();
   Ns=dfl.Ns;
   b0=blk_list(DFL_BLOCKS,&nb,&isw);
   bm=b0+nb;
   if (isw==0)
      b=b0;
   else
      b=b0+(nb/2);
   vm=v+nb*Ns;

   for (;v<vm;)
   {
      assign_sd2sdb(b,ALL_PTS,k,0);
      
      for (i=1;i<=Ns;i++)
      {
         z.re=-(*v).re;
         z.im=-(*v).im;
         mulc_spinor_add_dble((*b).vol,(*b).sd[0],(*b).sd[i],z);
         v+=1;
      }

      assign_sdb2sd(b,ALL_PTS,0,k);
      
      b+=1;
      if (b==bm)
         b=b0;
   }   
}


void dfl_s2v(int k,complex *v)
{
   int nb,isw,Ns,i;
   complex *vm;
   block_t *b0,*b,*bm;
   dfl_parms_t dfl;

   dfl=dfl_parms();
   Ns=dfl.Ns;
   b0=blk_list(DFL_BLOCKS,&nb,&isw);
   bm=b0+nb;
   if (isw==0)
      b=b0;
   else
      b=b0+(nb/2);
   vm=v+nb*Ns;

   for (;v<vm;)
   {
      assign_s2sb(b,ALL_PTS,k,0);
      
      for (i=1;i<=Ns;i++)
      {
         (*v)=spinor_prod((*b).vol,0,(*b).s[i],(*b).s[0]);
         v+=1;
      }

      b+=1;
      if (b==bm)
         b=b0;
   }   
}


void dfl_v2s(complex *v,int k)
{
   int nb,isw,Ns,i;
   complex *vm;
   block_t *b0,*b,*bm;
   dfl_parms_t dfl;

   dfl=dfl_parms();
   Ns=dfl.Ns;
   b0=blk_list(DFL_BLOCKS,&nb,&isw);
   bm=b0+nb;
   if (isw==0)
      b=b0;
   else
      b=b0+(nb/2);   
   vm=v+nb*Ns;

   for (;v<vm;)
   {
      set_s2zero((*b).vol,(*b).s[0]);
      
      for (i=1;i<=Ns;i++)
      {
         mulc_spinor_add((*b).vol,(*b).s[0],(*b).s[i],*v);
         v+=1;
      }

      assign_sb2s(b,ALL_PTS,0,k);
      
      b+=1;
      if (b==bm)
         b=b0;
   }   
}


void dfl_sub_v2s(complex *v,int k)
{
   int nb,isw,Ns,i;
   complex *vm,z;
   block_t *b0,*b,*bm;
   dfl_parms_t dfl;

   dfl=dfl_parms();
   Ns=dfl.Ns;
   b0=blk_list(DFL_BLOCKS,&nb,&isw);
   bm=b0+nb;
   if (isw==0)
      b=b0;
   else
      b=b0+(nb/2);    
   vm=v+nb*Ns;

   for (;v<vm;)
   {
      assign_s2sb(b,ALL_PTS,k,0);
      
      for (i=1;i<=Ns;i++)
      {
         z.re=-(*v).re;
         z.im=-(*v).im;
         mulc_spinor_add((*b).vol,(*b).s[0],(*b).s[i],z);
         v+=1;
      }

      assign_sb2s(b,ALL_PTS,0,k);
      
      b+=1;
      if (b==bm)
         b=b0;
   }   
}


static void assign_s2sdb(block_t *b,int k,int l)
{
   int i,*imb,*imm;
   float *r;
   double *rd;
   spinor *psk;

   imb=(*b).imb;
   imm=imb+(*b).vol;
   psk=ps[k][0];
   rd=(double*)((*b).sd[l]);

   for (;imb<imm;imb++)
   {
      r=(float*)(psk+(*imb));

      for (i=0;i<24;i++)
      {
         (*rd)=(double)(*r);
         r+=1;
         rd+=1;
      }
   }
}


void dfl_subspace(spinor **mds)
{
   int nb,isw;
   int ib,vol,i,j;
   complex *v,*vm;
   complex_dble z;
   spinor **s;
   spinor_dble **sd;
   block_t *b0,*b;

   alloc_modes();

   for (i=0;i<Nss;i++)
      ism[i]=(mds[i]-ps[0][0])/NSPIN;

   b0=blk_list(DFL_BLOCKS,&nb,&isw);
   if (isw==0)
      b=b0;
   else
      b=b0+(nb/2);
   
   for (ib=0;ib<nb;)
   {
      vol=(*b).vol;
      sd=(*b).sd;
      s=(*b).s;

      for (i=1;i<=Nss;i++)
      {
         assign_s2sdb(b,ism[i-1],i);
         v=vs[i-1]+ib*Nss;

         for (j=1;j<i;j++)
         {
            z=spinor_prod_dble(vol,0,sd[j],sd[i]);

            (*v).re=(float)(z.re);
            (*v).im=(float)(z.im);
            v+=1;
            
            z.re=-(z).re;
            z.im=-(z).im;
            mulc_spinor_add_dble(vol,sd[i],sd[j],z);
         }

         (*v).re=(float)(normalize_dble(vol,0,sd[i]));
         (*v).im=0.0f;
         v+=1;
         vm=v+(Nss-i);

         for (;v<vm;v++)
         {
            (*v).re=0.0f;
            (*v).im=0.0f;
         }
         
         assign_sd2s(vol,sd[i],s[i]);
      }

      b+=1;      
      ib+=1;

      if ((ib==(nb/2))&&(isw!=0))
         b=b0;
   }

   set_flags(NEW_DFLMODES);
}


void dfl_global_modes(complex **v)
{
   int i;

   for (i=0;i<Nss;i++)
      v[i]=vs[i];      
}
