
/*******************************************************************************
*
* File vflds.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Allocation of the global vector fields related to the deflation
* preconditioner
*
* The externally accessible functions are
*
*   void alloc_vflds(int n)
*     Allocates n global single-precision vector fields
*
*   void alloc_vdflds(int n)
*     Allocates n global double-precision vector fields
*
*   void free_vflds(void)
*     Frees the fields previously allocated by alloc_vflds()
*
*   void free_vdflds(void)
*     Frees the fields previously allocated by alloc_vdflds()
*
*   complex *vflds_base(int *n,int *nv,int *nvec)
*     Returns the address of the first single-precision vector field and
*     assigns the number of currently allocated fields to the parameter n.
*     The meaning of the parameters nv and nvec is explained in the notes.
*     If no fields are allocated, the program returns NULL
*
*   complex_dble *vdflds_base(int *n,int *nv,int *nvec)
*     Returns the address of the first double-precision vector field and
*     assigns the number of currently allocated fields to the parameter n.
*     The meaning of the parameters nv and nvec is explained in the notes.
*     If no fields are allocated, the program returns NULL
*
* Notes:
*
* The DFL_BLOCKS grid and the associated fields (index arrays and vector
* fields) can conveniently be allocated using the program set_dfl_parms()
* (see flags/parms.c). This is the recommended way to use the allocation
* programs in this file
*
* The fields are stored in a contiguous manner, where one field comes
* after the next. Each field has nvec complex components. The first nv
* of these are associated to the modes on the blocks in the local lattice,
* i.e. nv=nb*Ns, where nb and Ns are the numbers of blocks and fields per
* block, respectively, as set by the program set_dfl_parms(). The remaining
* nvec-nv components of each field are associated to the *even* blocks on
* the exterior boundary of the local lattice
*
* It is possible to access the fields using vflds_base() and vdflds_base(),
* noting that each field has exactly nvec components. Wherever possible, it
* is however recommended to access the fields via the workspace allocation
* programs (see flags/wspace.c)
* 
* Except for the programs vflds_base() and vdflds_base(), all programs
* in this module involve global communications and should therefore be
* called simultaneously on all processes
*
*******************************************************************************/

#define VFLDS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "block.h"
#include "flags.h"
#include "dfl.h"
#include "global.h"

static int no_v=0,no_vd=0,nvs,nvecs;
static complex *v=NULL;
static complex_dble *vd=NULL;


static void vfld_size(void)
{
   int nb,isw;
   int nbl[4];
   block_t *b;
   dfl_parms_t dfl;

   error_root(sizeof(complex)!=(2*sizeof(float)),1,
              "vfld_size [vflds.c]",
              "The complex structures are not properly packed");
   error_root(sizeof(complex_dble)!=(2*sizeof(double)),1,
              "vfld_size [vflds.c]",
              "The complex_dble structures are not properly packed");
   
   b=blk_list(DFL_BLOCKS,&nb,&isw);
   error(b==NULL,1,"vfld_size [vflds.c]",
         "The DFL_BLOCKS grid is not allocated");

   dfl=dfl_parms();
   error(dfl.Ns==0,1,"vfld_size [vflds.c]","The number of modes is not set\n"
         "(see set_dfl_parms() [flags/parms.c])");
   
   nbl[0]=L0/(*b).bs[0];
   nbl[1]=L1/(*b).bs[1];
   nbl[2]=L2/(*b).bs[2];
   nbl[3]=L3/(*b).bs[3];

   nvs=nb;
   nvecs=nb;

   if (NPROC0>1)
      nvecs+=nbl[1]*nbl[2]*nbl[3];
   if (NPROC1>1)
      nvecs+=nbl[0]*nbl[2]*nbl[3];
   if (NPROC2>1)
      nvecs+=nbl[0]*nbl[1]*nbl[3];
   if (NPROC3>1)
      nvecs+=nbl[0]*nbl[1]*nbl[2];
   
   nvs*=dfl.Ns;
   nvecs*=dfl.Ns;
}


void free_vflds(void)
{
   if (no_v!=0)
   {
      error(wv_size()!=0,1,"free_vflds [vflds.c]","Fields are in use");

      afree(v);
      v=NULL;
      no_v=0;

      reserve_wv(0);
   }
}


void free_vdflds(void)
{
   if (no_vd!=0)
   {
      error(wvd_size()!=0,1,"free_vdflds [vflds.c]","Fields are in use");

      afree(vd);
      vd=NULL;
      no_vd=0;

      reserve_wvd(0);
   }
}


void alloc_vflds(int n)
{
   int iprms[1];

   if (NPROC>1)
   {
      iprms[0]=n;
      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      error(iprms[0]!=n,1,"alloc_vflds [vflds.c]",
            "Parameter n is not global");
   }

   free_vflds();
   
   if (n>0)
   {
      vfld_size();
      v=amalloc(n*nvecs*sizeof(*v),ALIGN);

      error(v==NULL,1,"alloc_vflds [vflds.c]",
            "Unable to allocate vector fields");

      set_v2zero(n*nvecs,v);
      no_v=n;
   }
}


void alloc_vdflds(int n)
{
   int iprms[1];

   if (NPROC>1)
   {
      iprms[0]=n;
      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      error(iprms[0]!=n,1,"alloc_vdflds [vflds.c]",
            "Parameter n is not global");
   }

   free_vdflds();
   
   if (n>0)
   {
      vfld_size();
      vd=amalloc(n*nvecs*sizeof(*vd),ALIGN);

      error(vd==NULL,1,"alloc_vdflds [vflds.c]",
            "Unable to allocate vector fields");

      set_vd2zero(n*nvecs,vd);
      no_vd=n;
   }
}


complex *vflds_base(int *n,int *nv,int *nvec)
{
   *n=no_v;
   *nv=nvs;
   *nvec=nvecs;
   
   if (no_v>0)
      return v;
   else
      return NULL;
}
   

complex_dble *vdflds_base(int *n,int *nv,int *nvec)
{
   *n=no_vd;
   *nv=nvs;
   *nvec=nvecs;   
   
   if (no_vd>0)
      return vd;
   else
      return NULL;
}
