
/*******************************************************************************
*
* File vgrid.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Programs related to the geometry of the DFL_BLOCKS grid
*
* The externally accessible functions are
*
*   void vgrid(void)
*     Allocates and initializes the index arrays that describe the 
*     geometry of the currently allocated DFL_BLOCKS grid. This program
*     is called by the program that allocates (or reallocates) the grid,
*     so that it is not required to call it from anywhere else
*
*   int (*vinn(int *nb))[8]
*     Returns the starting address of the index array inn[][8] that
*     contains the indices of the nearest neighbours of the blocks in
*     the local lattice and assigns the number of the blocks to the
*     variable nb. The 8 directions -0,+0,..,-3,+3 are ordered in the
*     same way as the faces of the local lattice
*
*   int *vipp(int *nbbe,int *nbbo,int *ofsbbe,int *ofsbbo)
*     Returns the starting address of the array ipp[] of the indices of
*     the partner points of the exterior boundary points of the local
*     block lattice. The arguments nbbe,..,ofsbbo must be the starting
*     addresses of integer arrays with 8 elements. The program assigns
*     the numbers of the even and the odd points on the faces of the
*     lattice and the corresponding offsets to these
*   
* Notes:
*
* The blocks in the DFL_BLOCKS grid form a hypercubic lattice whose geometry
* can be described by suitable index arrays. When the blocks are allocated,
* they are ordered according to the local coordinates of their root points.
* First come all the even blocks and then all odd ones. Apart from that, the
* ordering of any two blocks with root points x and y is lexicographic, i.e. 
* such that
*
*   x<y if {(x0<y0) or ((x0=y0)&&(x1<y1)) or ...}
*
* This defines a total ordering of the blocks in the local lattices.
*
* While the number of the blocks in the local lattices must be even, the
* number of these blocks in any direction can be odd. As a result the even
* blocks may not be at even positions with respect to the root point of
* the total lattice
*
* The local labelling of the blocks defined by the program vgrid() divides
* the blocks into even and odd ones according to their position on the total
* lattice. Otherwise the ordering is the same as the local one. In other
* words, on the local lattices where the first block is even with respect to
* the total lattice, the ordering is exactly the same, while in the other
* cases the set of locally even blocks comes after the locally odd blocks
*
* On the exterior boundaries of the local block lattice, the same ordering
* is used as in the bulk, i.e. first come all even blocks then all odd ones 
* and both in lexicographic order, the reference point being the root point
* of the total lattice
*
* Explicitly the arrays returned by the programs vinn() and vipp() are
*
*  nb              Number of blocks in the local lattice
*
*  inn[ix][ifc]    Index of the nearest neighbour in direction ifc of
*                  the block with label ix (ix=0,..,nb-1, ifc=0,..,7).
*                  The ordering of the directions is -0,+0,..,-3,+3
*
*  ipp[ix]         Index of the nearest neighbour (partner point) in the
*                  local lattice of the block on the exterior boundary
*                  with index nb+ix (ix=0,..,nbb-1, nbb=number of blocks
*                  on the boundary)
*
*  nbbe[ifc]       Number of even (odd) blocks on the exterior boundary
*  nbbo[ifc]       in direction ifc
*
*  ofsbbe[ifc]     Offset of the index of the first even (odd) block on 
*  ofsbbo[ifc]     the exterior boundary in direction ifc
*
*******************************************************************************/

#define VGRID_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "start.h"
#include "block.h"
#include "global.h"

static int nbs,isws;
static int nbl[4],nbb[4],nbe[8],nbo[8],ofse[8],ofso[8];
static int *idx=NULL,*ipp=NULL,(*inn)[8]=NULL;


static void set_grid_sizes(void)
{
   int l[4],mu,ifc;
   block_t *b;

   b=blk_list(DFL_BLOCKS,&nbs,&isws);

   error(b==NULL,1,"set_grid_sizes [vgrid.c]",
         "DFL_BLOCKS grid is not allocated");

   l[0]=L0;
   l[1]=L1;
   l[2]=L2;
   l[3]=L3;

   for (mu=0;mu<4;mu++)
   {
      nbl[mu]=l[mu]/(*b).bs[mu];
      nbb[mu]=0;
   }

   if (NPROC0>1)
      nbb[0]=nbl[1]*nbl[2]*nbl[3];
   if (NPROC1>1)
      nbb[1]=nbl[0]*nbl[2]*nbl[3];
   if (NPROC2>1)
      nbb[2]=nbl[0]*nbl[1]*nbl[3];
   if (NPROC3>1)
      nbb[3]=nbl[0]*nbl[1]*nbl[2];

   for (mu=0;mu<4;mu++)
   {
      if (isws==1)
      {
         nbe[2*mu]=(nbb[mu]+1)/2;
         nbo[2*mu]=nbb[mu]-nbe[2*mu];
      }
      else
      {
         nbo[2*mu]=(nbb[mu]+1)/2;
         nbe[2*mu]=nbb[mu]-nbo[2*mu];
      }
      
      if ((nbl[mu]%2)==1)
      {
         nbe[2*mu+1]=nbe[2*mu];
         nbo[2*mu+1]=nbo[2*mu];
      }
      else
      {
         nbe[2*mu+1]=nbo[2*mu];
         nbo[2*mu+1]=nbe[2*mu];
      }
   }

   ofse[0]=0;
   
   for (ifc=1;ifc<8;ifc++)
      ofse[ifc]=ofse[ifc-1]+nbe[ifc-1];

   ofso[0]=ofse[7]+nbe[7];

   for (ifc=1;ifc<8;ifc++)
      ofso[ifc]=ofso[ifc-1]+nbo[ifc-1];   
}


static void check_block_positions(void)
{
   int nb,isw,*bs;
   int ie,n0,n1,n2,n3;
   block_t *bc[2],*b;
   
   bc[0]=blk_list(DFL_BLOCKS,&nb,&isw);
   bc[1]=bc[0]+(nb/2);

   bs=(*(bc[0])).bs;
   ie=0;

   for (n0=0;n0<nbl[0];n0++)
   {
      for (n1=0;n1<nbl[1];n1++)
      {
         for (n2=0;n2<nbl[2];n2++)
         {
            for (n3=0;n3<nbl[3];n3++)
            {
               if (((n0+n1+n2+n3)&0x1)==0)
               {
                  b=bc[0];
                  bc[0]+=1;
               }
               else
               {
                  b=bc[1];
                  bc[1]+=1;
               }
               
               if ((((*b).bo[0]-n0*bs[0])!=0)||
                   (((*b).bo[1]-n1*bs[1])!=0)||
                   (((*b).bo[2]-n2*bs[2])!=0)||
                   (((*b).bo[3]-n3*bs[3])!=0))
                  ie=1;
            }
         }
      }
   }

   error(ie!=0,1,"check_block_positions [vgrid.c]",
         "Unexpected block positions");
}


static void allocate_indices(void)
{
   if (idx!=NULL)
   {
      afree(idx);
      afree(inn);
   }

   idx=amalloc((nbs+2*(nbb[0]+nbb[1]+nbb[2]+nbb[3]))*sizeof(*idx),3);   
   inn=amalloc(nbs*sizeof(*inn),3);

   error((idx==NULL)||(inn==NULL),1,"allocate_indices [vgrid.c]",
         "Unable to allocate index arrays");

   ipp=idx+nbs;
}


static void set_index(void)
{
   int n0,n1,n2,n3;
   int ic[2],in;

   ic[0]=0;
   ic[1]=nbs/2;
   in=0;

   for (n0=0;n0<nbl[0];n0++)
   {
      for (n1=0;n1<nbl[1];n1++)
      {
         for (n2=0;n2<nbl[2];n2++)
         {
            for (n3=0;n3<nbl[3];n3++)
            {
               if (((n0+n1+n2+n3)&0x1)==isws)
               {
                  idx[in]=ic[0];
                  ic[0]+=1;
               }
               else
               {
                  idx[in]=ic[1];
                  ic[1]+=1;
               }

               in+=1;
            }
         }
      }
   }
}


static int index(int n0,int n1,int n2,int n3)
{
   n0=safe_mod(n0,nbl[0]);
   n1=safe_mod(n1,nbl[1]);
   n2=safe_mod(n2,nbl[2]);
   n3=safe_mod(n3,nbl[3]);   

   return idx[n3+nbl[3]*n2+nbl[2]*nbl[3]*n1+nbl[1]*nbl[2]*nbl[3]*n0];
}

   
void vgrid(void)
{
   int n0,n1,n2,n3,in;
   int ic[8],ifc,ip;

   set_grid_sizes();
   check_block_positions();
   allocate_indices();
   set_index();

   for (n0=0;n0<nbl[0];n0++)
   {
      for (n1=0;n1<nbl[1];n1++)
      {
         for (n2=0;n2<nbl[2];n2++)
         {
            for (n3=0;n3<nbl[3];n3++)
            {
               in=index(n0,n1,n2,n3);
               inn[in][0]=index(n0-1,n1,n2,n3);
               inn[in][1]=index(n0+1,n1,n2,n3);
               inn[in][2]=index(n0,n1-1,n2,n3);
               inn[in][3]=index(n0,n1+1,n2,n3);
               inn[in][4]=index(n0,n1,n2-1,n3);
               inn[in][5]=index(n0,n1,n2+1,n3);                 
               inn[in][6]=index(n0,n1,n2,n3-1);
               inn[in][7]=index(n0,n1,n2,n3+1);

               if (NPROC0>1)
               {
                  if (n0==0)
                     inn[in][0]=nbs;
                  if (n0==(nbl[0]-1))
                     inn[in][1]=nbs;
               }
               if (NPROC1>1)
               {               
                  if (n1==0)
                     inn[in][2]=nbs;
                  if (n1==(nbl[1]-1))
                     inn[in][3]=nbs;
               }
               if (NPROC2>1)
               {               
                  if (n2==0)
                     inn[in][4]=nbs;
                  if (n2==(nbl[2]-1))
                     inn[in][5]=nbs;
               }
               if (NPROC3>1)
               {               
                  if (n3==0)
                     inn[in][6]=nbs;
                  if (n3==(nbl[3]-1))
                     inn[in][7]=nbs;
               }
            }
         }
      }
   }

   for (ifc=0;ifc<8;ifc++)
      ic[ifc]=0;

   for (in=0;in<(nbs/2);in++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         if (inn[in][ifc]==nbs)
         {
            ip=ofso[ifc]+ic[ifc];
            ic[ifc]+=1;
            
            inn[in][ifc]=nbs+ip;
            ipp[ip]=in;
         }
      }
   }

   for (ifc=0;ifc<8;ifc++)
      ic[ifc]=0;

   for (in=(nbs/2);in<nbs;in++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         if (inn[in][ifc]==nbs)
         {
            ip=ofse[ifc]+ic[ifc];
            ic[ifc]+=1;
            
            inn[in][ifc]=nbs+ip;
            ipp[ip]=in;
         }
      }
   }   
}


int (*vinn(int *nb))[8]
{
   *nb=nbs;
      
   return inn;
}


int *vipp(int *nbbe,int *nbbo,int *ofsbbe,int *ofsbbo)
{
   int ifc;

   for (ifc=0;ifc<8;ifc++)
   {
      nbbe[ifc]=nbe[ifc];
      nbbo[ifc]=nbo[ifc];
      ofsbbe[ifc]=ofse[ifc];
      ofsbbo[ifc]=ofso[ifc];
   }

   return ipp;
}
