
/*******************************************************************************
*
* File vinit.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generic assignment and initialization programs for complex single- and 
* double-precision n-vectors
*
* The externally accessible functions are
*
*   void set_v2zero(int n,complex *v)
*     Sets the single-precision n-vector v[] to zero
*
*   void set_vd2zero(int n,complex_dble *v)
*     Sets the double-precision n-vector v[] to zero
*
*   void random_v(int n,complex *v,float sigma)
*     Initializes the components of the single-precision n-vector v[]
*     to (complex) random values z with distribution proportional to
*     exp{-|z|^2/sigma^2}
*
*   void random_vd(int n,complex_dble *v,double sigma)
*     Initializes the components of the double-precision n-vector v[]
*     to (complex) random values z with distribution proportional to
*     exp{-|z|^2/sigma^2}
*
*   void assign_v2v(int n,complex *v,complex *w)
*     Assigns the single-precision n-vector v to the single-precision
*     n-vector w[]
*
*   void assign_v2vd(int n,complex *v,complex_dble *w)
*     Assigns the single-precision n-vector v[] to the double-precision
*     n-vector w[]
*
*   void assign_vd2v(int n,complex_dble *v,complex *w)
*     Assigns the double-precision n-vector v[] to the single-precision
*     n-vector w[]
*
*   void assign_vd2vd(int n,complex_dble *v,complex_dble *w)
*     Assigns the double-precision n-vector v[] to the double-precision
*     n-vector w[]
*
* Notes:
*
* All these programs operate on arrays of complex fields, whose base 
* addresses are passed through the arguments. The length of the arrays 
* is specified by the parameter n
*
*******************************************************************************/

#define VINIT_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "su3.h"
#include "random.h"

static const complex v0={0.0f};
static const complex_dble vd0={0.0};


void set_v2zero(int n,complex *v)
{
   complex *vm;

   vm=v+n;
   
   for (;v<vm;v++)
      *v=v0;
}


void set_vd2zero(int n,complex_dble *v)
{
   complex_dble *vm;

   vm=v+n;
   
   for (;v<vm;v++)
      *v=vd0;
}

   
void random_v(int n,complex *v,float sigma)
{
   complex *vm;

   gauss((float*)(v),2*n);
   vm=v+n;
   
   for (;v<vm;v++)
   {
      (*v).re*=sigma;
      (*v).im*=sigma;
   }
}


void random_vd(int n,complex_dble *v,double sigma)
{
   complex_dble *vm;

   gauss_dble((double*)(v),2*n);
   vm=v+n;
   
   for (;v<vm;v++)
   {   
      (*v).re*=sigma;
      (*v).im*=sigma;
   }
}


void assign_v2v(int n,complex *v,complex *w)
{
   complex *vm;

   vm=v+n;
   
   for (;v<vm;v++)
   {
      (*w).re=(*v).re;
      (*w).im=(*v).im;
      w+=1;
   }
}


void assign_v2vd(int n,complex *v,complex_dble *w)
{
   complex *vm;

   vm=v+n;
   
   for (;v<vm;v++)
   {   
      (*w).re=(double)((*v).re);
      (*w).im=(double)((*v).im);
      w+=1;
   }
}


void assign_vd2v(int n,complex_dble *v,complex *w)
{
   complex_dble *vm;

   vm=v+n;
   
   for (;v<vm;v++)
   {   
      (*w).re=(float)((*v).re);
      (*w).im=(float)((*v).im);
      w+=1;
   }
}


void assign_vd2vd(int n,complex_dble *v,complex_dble *w)
{
   complex_dble *vm;

   vm=v+n;
   
   for (;v<vm;v++)
   {
      (*w).re=(*v).re;
      (*w).im=(*v).im;
      w+=1;
   }
}

