
/*******************************************************************************
*
* File force0.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of force that derives from the gauge action
*
* The externally accessible functions are
*
*   void hmc_frg(int n)
*     Computes the force on the n'th block of the HMC_BLOCKS grid, omitting
*     the prefactor 1/g0^2, and assigns the result to the force field
*
*   double hmc_acg(int n)
*     Returns the active part of the gauge-field action on the n'th block of
*     the HMC_BLOCKS grid. The active part of the action is equal to the sum 
*     of tr{1-U(p)} over all oriented plaquettes p that contain at least one
*     active link (the prefactor 1/g0^2 is thus omitted here too)
*
* Notes:
*
* The force is obtained by running through all double-plaquettes with corners
* labelled as 
*
*                 w *-------* z = x+mu+nu
*                   |       | 
*                   |       |
*                   |       |
*                 x *-------* y = x+mu
*                   |       | 
*                   |       |
*                   |       |
*                 s *-------* r = x+mu-nu
*
* where x is odd and mu<nu. The 7 links in these two plaquettes are
*
*  u0=U(x,y)    u2=U(x,w)    u4=U(r,y)    u6=U(s,r)      
*  u1=U(y,z)    u3=U(w,z)    u5=U(s,x)
*
* The corresponding index arrays are initialized when the HMC auxiliary
* fields are allocated
*
* Note that the force field is a shared field. Moreover the same field is
* used for the various forces (gauge, block, block interaction)
*
*******************************************************************************/

#define FORCE0_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "su3.h"
#include "start.h"
#include "misc.h"
#include "hmc.h"
#include "global.h"

#define MAX_LEVELS 8
#define BLK_LENGTH 4

#define _re(z,w) ((z).re*(w).re+(z).im*(w).im)

static int cnt[MAX_LEVELS];
static double smx[MAX_LEVELS];

#if (defined SSE2)
static su3_dble w1,w2 __attribute__ ((aligned (16)));
static su3_alg_dble X __attribute__ ((aligned (16)));
static const su3_alg_dble X0 __attribute__ ((aligned (16))) ={0.0};
#else
static su3_dble w1,w2;
static su3_alg_dble X;
static const su3_alg_dble X0={0.0};
#endif


static void set_frc2zero(hmc_t *hflds)
{
   int *lk,*lkm;
   su3_alg_dble *fb;

   fb=(*hflds).frc;   
   lk=(*hflds).lks;
   lkm=lk+(*hflds).nlks;
   
   for (;lk<lkm;lk++)
      *(fb+(*lk))=X0;
}


static void acc_frg(su3_dble *ub,hmc_t *hflds)
{
   int *pl,*plm;
   su3_alg_dble *fb,*fr;

   fb=(*hflds).frc;
   pl=(*hflds).plq;
   plm=pl+14*(*hflds).nplq;

   for (;pl<plm;pl+=14)
   {
      if ((pl[7]+pl[8]+pl[9]+pl[10])!=0)
      {
         su3xsu3(ub+pl[0],ub+pl[1],&w1);
         su3dagxsu3dag(ub+pl[3],ub+pl[2],&w2);

         if ((pl[7]+pl[9])!=0)
         {
            prod2su3alg(&w1,&w2,&X);            

            if (pl[7]!=0)
            {
               fr=fb+pl[0];
               _su3_alg_add_assign(*fr,X);
            }

            if (pl[9]!=0)
            {
               fr=fb+pl[2];
               _su3_alg_sub_assign(*fr,X);
            }            
         }

         if ((pl[8]+pl[10])!=0)
         {
            prod2su3alg(&w2,&w1,&X);               

            if (pl[8]!=0)
            {
               fr=fb+pl[1];
               _su3_alg_add_assign(*fr,X);
            }

            if (pl[10]!=0)
            {
               fr=fb+pl[3];
               _su3_alg_sub_assign(*fr,X);
            }            
         }         
      }
      
      if ((pl[7]+pl[11]+pl[12]+pl[13])!=0)
      {
         su3xsu3dag(ub+pl[0],ub+pl[4],&w1);
         su3dagxsu3(ub+pl[6],ub+pl[5],&w2);

         if ((pl[7]+pl[12])!=0)
         {
            prod2su3alg(&w1,&w2,&X);            

            if (pl[7]!=0)
            {
               fr=fb+pl[0];
               _su3_alg_add_assign(*fr,X);
            }

            if (pl[12]!=0)
            {
               fr=fb+pl[5];
               _su3_alg_add_assign(*fr,X);
            }            
         }

         if ((pl[11]+pl[13])!=0)
         {
            prod2su3alg(&w2,&w1,&X);                        

            if (pl[11]!=0)
            {
               fr=fb+pl[4];
               _su3_alg_sub_assign(*fr,X);
            }

            if (pl[13]!=0)
            {
               fr=fb+pl[6];
               _su3_alg_sub_assign(*fr,X);
            }            
         }         
      }
   }
}


static void cnv_frg(su3_dble *ub,hmc_t *hflds)
{
   int ilk,*lk,*lkm;
   su3_alg_dble *fb,*fr,*frn;
   su3_dble *u;

   fb=(*hflds).frc;
   frn=fb;
   lk=(*hflds).lks;
   lkm=lk+(*hflds).nlks;
   
   for (;lk<lkm;lk++)
   {
      ilk=(*lk);
      u=ub+ilk;
      fr=fb+ilk;
      
      if ((ilk&0x1)!=0)
         rotate_su3alg(u,fr);

      *frn=*fr;
      frn+=1;
   }
}


void hmc_frg(int n)
{
   int nb,isw;
   su3_dble *ub;
   block_t *b;
   hmc_t *hflds;

   b=blk_list(HMC_BLOCKS,&nb,&isw)+n;
   hflds=hmcflds()+n;
   ub=(*b).ud[0];
      
   set_frc2zero(hflds);
   acc_frg(ub,hflds);
   cnv_frg(ub,hflds);
}


static double act_plq(int *pl,int *plm,su3_dble *ub)
{
   int n,*pln;
   double act,ds;
   
   for (n=0;n<MAX_LEVELS;n++)
   {
      cnt[n]=0;
      smx[n]=0.0;
   }

   for (;pl<plm;)
   {
      act=0.0;
      pln=pl+BLK_LENGTH*14;

      for (;pl<pln;pl+=14)
      {
         if ((pl[7]+pl[8]+pl[9]+pl[10])!=0)
         {
            su3xsu3(ub+pl[0],ub+pl[1],&w1);
            su3xsu3(ub+pl[2],ub+pl[3],&w2);         

            ds=_re(w1.c11,w2.c11)+_re(w1.c12,w2.c12)+_re(w1.c13,w2.c13)+
               _re(w1.c21,w2.c21)+_re(w1.c22,w2.c22)+_re(w1.c23,w2.c23)+
               _re(w1.c31,w2.c31)+_re(w1.c32,w2.c32)+_re(w1.c33,w2.c33);
            act+=(3.0-ds);
         }

         if ((pl[7]+pl[11]+pl[12]+pl[13])!=0)
         {
            su3xsu3(ub+pl[6],ub+pl[4],&w1);
            su3xsu3(ub+pl[5],ub+pl[0],&w2);         

            ds=_re(w1.c11,w2.c11)+_re(w1.c12,w2.c12)+_re(w1.c13,w2.c13)+
               _re(w1.c21,w2.c21)+_re(w1.c22,w2.c22)+_re(w1.c23,w2.c23)+
               _re(w1.c31,w2.c31)+_re(w1.c32,w2.c32)+_re(w1.c33,w2.c33);
            act+=(3.0-ds);
         }
      }

      cnt[0]+=1;
      smx[0]+=act;

      for (n=1;(cnt[n-1]>=BLK_LENGTH)&&(n<MAX_LEVELS);n++)
      {
         cnt[n]+=1;
         smx[n]+=smx[n-1];

         cnt[n-1]=0;
         smx[n-1]=0.0;
      }         
   }

   act=0.0;

   for (n=0;n<MAX_LEVELS;n++)
      act+=smx[n]; 
   
   return act+act;
}


double hmc_acg(int n)
{
   int *pl,*plm,nb,isw;
   block_t *b;
   hmc_t *hflds;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   if (error_loc((b==NULL)||(hflds==NULL)||(n<0)||(n>=nb),1,
                 "hmc_acg [force0.c]",
                 "Improper block grid, block fields or block number"))
      return 0.0;

   b+=n;
   hflds+=n;
   pl=(*hflds).plq;
   plm=pl+14*(*hflds).nplq;

   return act_plq(pl,plm,(*b).ud[0]);
}

