
/*******************************************************************************
*
* File fgcr4vd.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generic flexible GCR inversion program for operators Op acting on
* double-precision complex fields
*
* The externally accessible function defined in this file is
*
*   double fgcr4vd(int vol,int icom,
*                  void (*Op)(complex_dble *v,complex_dble *w),
*                  void (*Mj)(int j,complex *r,complex *v,complex *opv),
*                  complex *wv[],complex_dble *wvd[],int nkv,int nmx,
*                  double res,complex_dble *v,complex_dble *w,int *status)
*     Solves the linear equation Op*w[]=v[] for given source field v[]. The
*     parameters vol, nkv, nmx and res denote the number of components in each
*     field, the number of Krylov vectors to be generated before a restart,
*     the maximal total number of Krylov vectors and the desired reduction of
*     the residue. The program operates globally if (and only if) icom=1.
*      A workspace of 2*nkv+1 single- and 1 double-precision fields is needed
*     that are pointed to by the arrays wv[] and wvd[]. On exit status reports
*     the total number of Krylov vectors that were used, or a negative value
*     if the program failed. The return value is the norm of the last residue
*
* Notes:
*
* This program uses single-precision arithmetic in order to reduce the 
* execution time but obtains the solution with double-precision accuracy
*
* The programs for the operator and the preconditioner must be externally
* defined and are assumed to have the following properties:
*
*   void Op(complex_dble *v,complex_dble *w)
*     Application of the operator Op to the double-precision field v[]
*     and assignement of the result to w[]. On exit the field v[] is
*     unchanged
*
*   void Mj(int j,complex *r,complex *v,complex *opv)
*     Application of the preconditioner M_j to the current residue r[] at
*     step j of the inner loop of the algorithm. On output r[] is unchanged,
*     while v[] and opv[] contain the new search direction v and the field 
*     Op*v respectively
*
*******************************************************************************/

#define FGCR4VD_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "linalg.h"
#include "dfl.h"
#include "global.h"

#define PRECISION_LIMIT ((double)(100.0f*FLT_EPSILON))

static int nvs,com,nkvm=0;
static float *be;
static double rn;
static complex *al,*ga,*ca;
static complex **psq,**psz,*psr;
static complex_dble *pdb,*pdx,*pdw,*cs1,*cs2;


static int alloc_arrays(int nkv)
{
   if (nkvm>0)
   {
      afree(al);
      afree(be);
      afree(cs1);
   }

   al=amalloc(nkv*(nkv+2)*sizeof(complex),3);
   be=amalloc(nkv*sizeof(float),3);
   cs1=amalloc((2*nkv+4)*sizeof(complex_dble),3);

   if ((al==NULL)||(be==NULL)||(cs1==NULL))
      return 1;

   ga=al+nkv*nkv;
   ca=ga+nkv;
   cs2=cs1+nkv+2;
   nkvm=nkv;

   return 0;
}


static void fgcr_init(int vol,int icom,int nkv,complex *wv[],
                      complex_dble *wvd[],complex_dble *v,complex_dble *w)
{
   nvs=vol;
   com=icom;

   psz=wv;
   psr=wv[nkv];
   psq=wv+nkv+1;

   pdb=v;
   pdx=w;
   pdw=wvd[0];

   set_vd2zero(nvs,pdx);
   assign_vd2v(nvs,pdb,psr);

   rn=(double)(vnorm_square(nvs,com,psr));
   rn=sqrt(rn);
}


static void fgcr_reset(void (*Op)(complex_dble *v,complex_dble *w))
{
   complex *rsr,*rsm;
   complex_dble *rdb,*rdw;

   (*Op)(pdx,pdw);

   rdb=pdb;
   rdw=pdw;
   rsm=psr+nvs;

   for (rsr=psr;rsr<rsm;rsr++)
   {
      (*rsr).re=(float)((*rdb).re-(*rdw).re);
      (*rsr).im=(float)((*rdb).im-(*rdw).im);      

      rdb+=1;
      rdw+=1;
   }

   rn=(double)(vnorm_square(nvs,com,psr));
   rn=sqrt(rn);
}


static void sum_vprod(int n)
{
   int i;
   
   if ((com==1)&&(NPROC>1))
   {
      MPI_Reduce((double*)(cs1),(double*)(cs2),2*n,MPI_DOUBLE,
                 MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast((double*)(cs2),2*n,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
   else
   {
      for (i=0;i<n;i++)
      {
         cs2[i].re=cs1[i].re;
         cs2[i].im=cs1[i].im;
      }
   }
}
   

static void scale_v(int n,complex *v,float r)
{
   complex *vm;

   vm=v+n;
   
   for (;v<vm;v++)
   {
      (*v).re*=r;
      (*v).im*=r;      
   }
}


static void next_zqr(void (*Mj)(int j,complex *r,complex *v,complex *opv),
                     int j,int nkv)
{
   int i;
   double r1,r2;
   complex *psw,*z,w;

   psw=psq[j];
   (*Mj)(j,psr,psz[j],psw);

   for (i=0;i<j;i++)
   {
      w=vprod(nvs,0,psq[i],psw);
      cs1[i].re=(double)(w.re);
      cs1[i].im=(double)(w.im);
   }

   sum_vprod(j);
   z=al+j;
   
   for (i=0;i<j;i++)
   {
      (*z).re=(float)(cs2[i].re);
      (*z).im=(float)(cs2[i].im);
      w.re=-(*z).re;
      w.im=-(*z).im;
      mulc_vadd(nvs,psw,psq[i],w);
      z+=nkv;
   }

   w.re=vnorm_square(nvs,0,psw);
   cs1[0].re=(double)(w.re);
   cs1[0].im=0.0;
   w=vprod(nvs,0,psw,psr);
   cs1[1].re=(double)(w.re);
   cs1[1].im=(double)(w.im);

   sum_vprod(2);
   r1=sqrt(cs2[0].re);
   be[j]=(float)(r1);

   if (r1==0.0)
      return;

   r2=1.0/r1;
   scale_v(nvs,psw,(float)(r2));

   z=ga+j;
   (*z).re=(float)(cs2[1].re*r2);
   (*z).im=(float)(cs2[1].im*r2);   
   w.re=-(*z).re;
   w.im=-(*z).im;
   mulc_vadd(nvs,psr,psw,w);

   rn=(double)(vnorm_square(nvs,com,psr));
   rn=sqrt(rn);
}


static void add_to_x(int nkv,int n)
{
   int k,j,kj;
   float r;
   complex *rsx,*rsm,sm;
   complex_dble *rdx;

   set_v2zero(nvs,psr);

   for (k=(n-1);k>=0;k--)
   {
      sm.re=ga[k].re;
      sm.im=ga[k].im;

      for (j=(k+1);j<n;j++)
      {
         kj=k*nkv+j;
         sm.re-=(al[kj].re*ca[j].re-al[kj].im*ca[j].im);
         sm.im-=(al[kj].re*ca[j].im+al[kj].im*ca[j].re);
      }

      r=1.0f/be[k];
      ca[k].re=sm.re*r;
      ca[k].im=sm.im*r;

      mulc_vadd(nvs,psr,psz[k],ca[k]);
   }

   rdx=pdx;
   rsm=psr+nvs;

   for (rsx=psr;rsx<rsm;rsx++)
   {
      (*rdx).re+=(double)((*rsx).re);
      (*rdx).im+=(double)((*rsx).im);

      rdx+=1;
   }
}


double fgcr4vd(int vol,int icom,void (*Op)(complex_dble *v,complex_dble *w),
               void (*Mj)(int j,complex *r,complex *v,complex *opv),
               complex *wv[],complex_dble *wvd[],int nkv,int nmx,
               double res,complex_dble *v,complex_dble *w,int *status)
{
   int n,j,iprms[3];
   double rn_old,tol,dprms[1];

   if ((icom==1)&&(NPROC>1))
   {
      iprms[0]=vol;
      iprms[1]=nkv;
      iprms[2]=nmx;
      dprms[0]=res;

      MPI_Bcast(iprms,3,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(dprms,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((iprms[0]!=vol)||(iprms[1]!=nkv)||(iprms[2]!=nmx)||
            (dprms[0]!=res),1,"fgcr4vd [fgcr4vd.c]",
            "Parameters are not global");

      error_root((vol<=0)||(nkv<1)||(nmx<1)||(res<=DBL_EPSILON),1,
                 "fgcr4vd [fgcr4vd.c]","Improper choice of vol,nkv,nmx or res");
      
      if (nkv>nkvm)
         error(alloc_arrays(nkv),1,"fgcr4vd [fgcr4vd.c]",
               "Unable to allocate auxiliary arrays");
   }
   else if (error_loc((vol<=0)||(nkv<1)||(nmx<1)||(res<=DBL_EPSILON),1,
                      "fgcr4vd [fgcr4vd.c]",
                      "Improper choice of vol,nkv,nmx or res"))
   {
      (*status)=0;
      return 1.0;
   }
   else if ((nkv>nkvm)&&(error_loc(alloc_arrays(nkv),1,"fgcr4vd [fgcr4vd.c]",
                                   "Unable to allocate auxiliary arrays")))
   {
      (*status)=0;
      return 1.0;
   }

   fgcr_init(vol,icom,nkv,wv,wvd,v,w);
   tol=res*rn;
   (*status)=0;

   while (rn>tol)
   {
      /* message("rn_old = %.2e\n",rn); */

      rn_old=rn;
      n=0;

      for (j=0;j<nkv;j++)
      {
         next_zqr(Mj,j,nkv);
         (*status)+=1;

         /* message("j = %d, rn = %.2e\n",j,rn); */
         
         if (be[j]!=0.0f)
         {
            n+=1;
            if ((rn<=tol)||(rn<(PRECISION_LIMIT*rn_old))||((*status)>=nmx))
               break;
         }
         else
            break;
      }

      add_to_x(nkv,n);
      fgcr_reset(Op);

      if (((*status)>=nmx)&&(rn>tol))
      {
         (*status)=-1;
         break;
      }
   }

   return rn;
}
