
/*******************************************************************************
*
* File mutils.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Utility functions used in main programs
*
* The externally accessible functions are
*
*   int find_opt(int argc,char *argv[],char *opt)
*     On process 0, this program compares the string opt with the arguments
*     argv[1],..,argv[argc-1] and returns the position of the first argument
*     that matches the string. If there is no matching argument, or if the
*     program is called from another process, the return value is 0
*
*   void check_dir(char* dir)
*     This program checks whether the directory dir is locally accessible,
*     from each process, and aborts the main program with an informative
*     error message if this is not the case. The program must be called
*     simultaneously on all processes, but the argument may depend on the
*     process
*
*   void check_dir_root(char* dir)
*     On process 0, this program checks whether the directory dir is
*     accessible and aborts the main program with an informative error
*     message if this is not the case. When called on other processes,
*     the program does nothing
*
*   int name_size(char *format,...)
*     On process 0, this program returns the length of the string that
*     would be printed by calling sprintf(*,format,...). The format
*     string can be any combination of literal text and the conversion
*     specifiers %s, %d and %.nf (where n is a positive integer). When
*     called on other processes, the program does nothing and returns
*     the value of NAME_SIZE
*     
*   void read_line(char *tag,char *format,...)
*     On process 0, this program reads a line of text and data from stdin
*     in a controlled manner, as described in the notes below. When
*     called on other processes, the program does nothing
*
* Notes:
*
* Except for check_dir(), the programs in this module do not involve any
* communications and can be called locally
*
* The program read_line() serves to read structured input parameter
* files (such as main/run3.in). It is assumed that the file was opened
* for formatted read on the stream stdin and that each line on the
* file has the form
*
*   tag v1 v2 ...
*
* where tag is a possibly empty string and v1,v2,... are data values
* (strings, integers or floating-point numbers) separated by blanks
*
* Such a line is read by calling
*
*   read_line(tag,format,&var1,&var2,...)
*
* where var1,var2,... are the variables to which the values v1,v2,... are
* to be assigned. The format string must include the associated sequence
* of conversion specifiers %s, %d, %f or %lf without any modifiers. Other
* tokens are not allowed in the format string, except for blanks and a
* newline character at the end of the string (none of these are required,
* but one may wish to to insert them for readability)
*
* When a line is read, the program performs various checks and aborts the
* main program with an informative error message if an inconsistency (tag
* mismatch, wrong read count, malformed format string) is encountered. The
* program also stops if a data string longer than NAME_SIZE-1 characters
* is to be read. Buffer overflows are thus excluded if the target string
* variables are allocated with size>=NAME_SIZE
*
*******************************************************************************/

#define MUTILS_C

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include "mpi.h"
#include "start.h"
#include "global.h"

static char line[NAME_SIZE+1],*sfmt=NULL;
static char inum[3*sizeof(int)+4];


int find_opt(int argc,char *argv[],char *opt)
{
   int my_rank,k;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      for (k=1;k<argc;k++)
         if (strcmp(argv[k],opt)==0)
            return k;
   }

   return 0;
}


void check_dir(char* dir)
{
   int my_rank,n;
   char *tmp_file;
   FILE *tmp;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   n=strlen(dir);
   tmp_file=malloc(n+8+3*sizeof(int));
   error(tmp_file==NULL,1,"check_dir [mutils.c]",
         "Unable to allocate name string");
   sprintf(tmp_file,"%s/.tmp_%x",dir,my_rank);

   n=0;
   tmp=fopen(tmp_file,"rb");

   if (tmp==NULL)
   {
      n=1;
      tmp=fopen(tmp_file,"wb");
   }

   error_loc(tmp==NULL,1,"check_dir [mutils.c]",
             "Unable to access directory %s",dir);
   error_chk();

   fclose(tmp);
   if (n==1)
      remove(tmp_file);
   free(tmp_file);
}


void check_dir_root(char* dir)
{
   int my_rank,n;
   char *tmp_file;
   FILE *tmp;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      n=strlen(dir);
      tmp_file=malloc((n+6)*sizeof(char));
      error_root(tmp_file==NULL,1,"check_dir_root [mutils.c]",
                 "Unable to allocate name string");
      sprintf(tmp_file,"%s/.tmp",dir);

      n=0;
      tmp=fopen(tmp_file,"rb");

      if (tmp==NULL)
      {
         n=1;
         tmp=fopen(tmp_file,"wb");
      }

      error_root(tmp==NULL,1,"check_dir_root [mutils.c]",
                 "Unable to access directory %s from process 0",dir);

      fclose(tmp);
      if (n==1)
         remove(tmp_file);
      free(tmp_file);
   }
}


int name_size(char *format,...)
{
   int my_rank,nlen,ie,n;
   double dmy;
   char *pp,*pc;
   va_list args;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      va_start(args,format);      
      pc=format;
      nlen=strlen(format);
      ie=0;
      n=0;
      
      for (;;)
      {
         pp=strchr(pc,'%');

         if (pp==NULL)
            break;

         pc=pp+1;
         
         if (pc[0]=='s')
            nlen+=(strlen(va_arg(args,char*))-2);
         else if (pc[0]=='d')
         {
            sprintf(inum,"%d",va_arg(args,int));
            nlen+=(strlen(inum)-2);
         }
         else if (pc[0]=='.')
         {
            if (sscanf(pc,".%d",&n)!=1)
            {
               ie=1;
               break;
            }
            
            sprintf(inum,".%df",n);
            pp=strstr(pc,inum);
            
            if (pp!=pc)
            {
               ie=2;
               break;
            }

            nlen+=(n+1-strlen(inum));
            dmy=va_arg(args,double);
            if (dmy<0.0)
               nlen+=1;
         }
         else
         {
            ie=3;
            break;
         }
      }

      va_end(args);
      error_root(ie!=0,1,"name_size [mutils.c]",
                 "Incorrect format string %s (ie=%d)",format,ie);
      return nlen;
   }

   return NAME_SIZE;
}


void read_line(char *tag,char *format,...)
{
   int my_rank,ndig,ic,ie;
   char *p,*s;
   va_list args;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      if (sfmt==NULL)
      {
         ndig=(int)(1.0+log10((double)(NAME_SIZE+1)));         
         sfmt=malloc((ndig+4)*sizeof(char));
         error_root(sfmt==NULL,1,"read_line [mutils.c]",
                    "Unable to allocate auxiliary string");
         sprintf(sfmt," %%%ds",NAME_SIZE);
      }

      if (tag[0]!='\0')
      {
         ic=scanf(sfmt,line);
         error_root(ic!=1,1,"read_line [mutils.c]",
                    "Missing line tag (expected %s)",tag);
         error_root(strlen(line)>=NAME_SIZE,1,"read_line [mutils.c]",
                    "Line tag %.24s ... read from input stream appears\n"
                    "to be longer than NAME_SIZE-1 (expected %s)",line,tag);
         error_root(strcmp(tag,line)!=0,1,"read_line [mutils.c]",
                    "Line tag %s is incorrect (expected %s)",line,tag);
      }
      
      va_start(args,format);      
      
      for (p=format;;p++)
      {
         p+=strspn(p," ");
         ie=0;
         ic=0;

         if (p[0]=='\0')
            break;         
         else if (p[0]=='\n')
         {
            scanf("\n");
            break;
         }
         else if (p[0]=='%')
         {
            p+=1;
            if (p[0]=='s')
            {
               s=va_arg(args,char*);
               ic=scanf(sfmt,line);
               if (ic==1)
               {
                  error_root(strlen(line)>=NAME_SIZE,1,"read_line [mutils.c]",
                             "Data string %.24s ... on line with tag %s\n"
                             "appears to be longer than NAME_SIZE-1",line,tag);
                  strcpy(s,line);
               }
            }
            else if (p[0]=='d')
               ic=scanf(" %d",va_arg(args,int*));
            else if (p[0]=='f')
               ic=scanf(" %f",va_arg(args,float*));
            else if (p[0]=='l')
            {
               p+=1;
               if (p[0]=='f')
                  ic=scanf(" %lf",va_arg(args,double*));
               else
                  ie=1;
            }
            else
               ie=1;
         }
         else
            ie=1;

         error_root(ie==1,1,"read_line [mutils.c]",
                    "Incorrect format string %s on line with tag %s",
                    format,tag);
         error_root(ic!=1,1,"read_line [mutils.c]",
                    "Missing data item(s) on line with tag %s",tag);  
      }

      va_end(args);
   }
}
