
/*******************************************************************************
*
* File update.c
*
* Copyright (C) 2006 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Update program for the pure SU(3) gauge theory with Wilson plaquette action
*
* The externally accessible functions are
*
*   link_t *active_links(int *nlk)
*     Determines the set of active links on the local lattice. The program
*     first calculates the number nlk of the active links and allocates an
*     array of links with nlk elements. The array is then initialized with
*     the sequence of active links and its address is returned
*
*   void update(float beta,int nhb,int nor,int rnfreq)
*     Updates the active single-precision link variables at inverse coupling
*     beta=6/g_0^2 by performing nhb heatbath sweeps followed by nor over-
*     relaxation sweeps. When the internal sweep count equals rnfreq, or if
*     nhb=nor=0, the gauge field is renormalized and the counter is reset to
*     zero. Before any update sweeps are performed, the field is translated
*     by a random vector in a direction orthogonal to the boundaries of the
*     local lattice
*
* Notes:
*
* Both of these programs operate globally and should therefore be called
* simultaneously on all processes.
*
* An active link is, by definition, such that the staples can be computed
* using local data only. These are precisely all those links on the local
* lattice that have at most one endpoint on the set of interior boundary
* points.
*
* It is advisable to set the renormalization frequency rnfreq to values
* not much larger than 10.
*
* The translation vectors have random components equal to 1 or 2 in the
* directions where the number of processors is greater than 1. All other
* components are set to zero.
*
*******************************************************************************/

#define UPDATE_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "su3.h"
#include "start.h"
#include "flags.h"
#include "misc.h"
#include "update.h"
#include "global.h"

static int nlks,rncnt=0;
static link_t *lks=NULL;

#if (defined SSE)
static su3 vmat __attribute__ ((aligned (16)));
#else
static su3 vmat;
#endif


link_t *active_links(int *nlk)
{
   int bs[4],ix,iy,mu,nu,il;
   int itx,ity;
   link_t *lk;

   bs[0]=L0-2*(1-(NPROC0%2));
   bs[1]=L1-2*(1-(NPROC1%2));
   bs[2]=L2-2*(1-(NPROC2%2));
   bs[3]=L3-2*(1-(NPROC3%2));

   (*nlk)=4*bs[0]*bs[1]*bs[2]*bs[3];

   if (NPROC0>1)
      (*nlk)+=bs[1]*bs[2]*bs[3];
   if (NPROC1>1)
      (*nlk)+=bs[2]*bs[3]*bs[0];
   if (NPROC2>1)
      (*nlk)+=bs[3]*bs[0]*bs[1];
   if (NPROC3>1)
      (*nlk)+=bs[0]*bs[1]*bs[2];

   lk=amalloc((*nlk)*sizeof(link_t),3);
   error(lk==NULL,1,"active_links [update.c]",
         "Unable to allocate index array");

   il=0;

   for (ix=0;ix<VOLUME;ix++)
   {
      for (mu=0;mu<4;mu++)
      {
         iy=iup[ix][mu];

         if (iy<VOLUME)
         {
            itx=0;
            ity=0;

            for (nu=0;nu<4;nu++)
            {
               if ((iup[ix][nu]>=VOLUME)||(idn[ix][nu]>=VOLUME))
                  itx=1;
               if ((iup[iy][nu]>=VOLUME)||(idn[iy][nu]>=VOLUME))
                  ity=1;
            }

            if ((itx==0)||(ity==0))
            {
               lk[il].ix=ix;
               lk[il].mu=mu;
               il+=1;
            }
         }
      }
   }

   return lk;
}


static void shift_field(void)
{
   int np[4],s[4],mu;
   double r[4];

   np[0]=NPROC0;
   np[1]=NPROC1;
   np[2]=NPROC2;
   np[3]=NPROC3;

   ranlxd(r,4);

   for (mu=0;mu<4;mu++)
   {
      if (np[mu]>1)
         s[mu]=(int)(2.0*r[mu])+1;
      else
         s[mu]=0;
   }
   
   MPI_Bcast(s,4,MPI_INT,0,MPI_COMM_WORLD);

   shift_u(s);
}


static void upd_active(float beta,int type,int rnfreq)
{
   int ix,mu;
   link_t *lk,*lm;

   lk=lks;
   lm=lk+nlks;

   for (;lk<lm;lk++)
   {
      ix=(*lk).ix;
      mu=(*lk).mu;

      staples(ix,mu,&vmat);
      cabmar(beta,pu[ix][mu],&vmat,type);
   }

   rncnt+=1;

   if (rncnt>=rnfreq)
   {
      renormalize_u();
      rncnt=0;
   }
}


void update(float beta,int nhb,int nor,int rnfreq)
{
   int n,iprms[3];
   float fprms[1];

   if (NPROC>1)
   {
      iprms[0]=nhb;
      iprms[1]=nor;
      iprms[2]=rnfreq;
      fprms[0]=beta;

      MPI_Bcast(iprms,3,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(fprms,1,MPI_FLOAT,0,MPI_COMM_WORLD);

      error((iprms[0]!=nhb)||(iprms[1]!=nor)||(iprms[2]!=rnfreq)||
            (fprms[0]!=beta),1,"update [update.c]",
            "Input parameters are not global");
   }
   
   if (lks==NULL)
      lks=active_links(&nlks);

   if ((nhb==0)&&(nor==0))
   {
      renormalize_u();
      rncnt=0;
   }

   if (((nhb>0)||(nor>0))&&(NPROC>1))
      shift_field();

   for (n=0;n<nhb;n++)
      upd_active(beta,0,rnfreq);

   for (n=0;n<nor;n++)
      upd_active(beta,1,rnfreq);

   set_flags(NEW_U);
}

